__all__ = ("MassFuncPress74",)

import numpy as np

from . import MassFunc, get_delta_c


class MassFuncPress74(MassFunc):
    """Implements the mass function of `Press & Schechter 1974
    <https://ui.adsabs.harvard.edu/abs/1974ApJ...187..425P/abstract>`_.
    This parametrization is only valid for 'fof' masses.

    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef` or :obj:`str`):
            a mass definition object, or a name string.
        mass_def_strict (:obj:`bool`): if ``False``, consistency of the mass
            definition will be ignored.
    """
    name = 'Press74'

    def __init__(self, *,
                 mass_def="fof",
                 mass_def_strict=True):
        super().__init__(mass_def=mass_def, mass_def_strict=mass_def_strict)
        self._norm = np.sqrt(2/np.pi)

    def _check_mass_def_strict(self, mass_def):
        return mass_def.Delta != "fof"

    def _get_fsigma(self, cosmo, sigM, a, lnM):
        delta_c = get_delta_c(cosmo, a, kind='EdS')
        nu = delta_c/sigM
        return self._norm * nu * np.exp(-0.5 * nu**2)



import pytest
import numpy as np
import pyccl as ccl

COSMO = ccl.Cosmology(Omega_c=0.25, Omega_b=0.05, Omega_g=0, Omega_k=0,
                      h=0.7, sigma8=0.8, n_s=0.96, Neff=0, m_nu=0.0,
                      w0=-1, wa=0, T_CMB=2.7)


def test_mdef_eq():
    hmd_200m = ccl.halos.MassDef200m
    hmd_200m_b = ccl.halos.MassDef(200, 'matter')
    assert hmd_200m == hmd_200m_b


def test_concentration_translation():
    c_old = np.array([9., 10., 11.])
    Delta_old = 200.

    # No change expected
    Delta_new = 200.
    c_new = ccl.halos.massdef.convert_concentration(
        COSMO, c_old=c_old, Delta_old=Delta_old, Delta_new=Delta_new)
    assert np.all(c_old == c_new)

    # Test against numerical solutions from Mathematica.
    Delta_new = 500.
    c_new = ccl.halos.massdef.convert_concentration(
        COSMO, c_old=c_old, Delta_old=Delta_old, Delta_new=Delta_new)
    c_new_expected = np.array([6.12194, 6.82951, 7.53797])
    assert np.all(np.fabs(c_new/c_new_expected-1) < 1E-4)


def test_init_raises():
    with pytest.raises(ValueError):
        ccl.halos.MassDef('bir', 'matter')

    with pytest.raises(ValueError):
        ccl.halos.MassDef('vir', 'radiation')

    with pytest.raises(ValueError):
        ccl.halos.MassDef(-100, 'matter')


def test_get_Delta():
    for Delta in [200, 500, 121]:
        hmd = ccl.halos.MassDef(Delta, 'critical')
        assert int(hmd.get_Delta(COSMO, 1.)) == Delta

    hmd = ccl.halos.MassDef('vir', 'critical')
    assert np.isfinite(hmd.get_Delta(COSMO, 1.))

    hmd = ccl.halos.MassDef('fof', 'critical')
    with pytest.raises(ValueError):
        hmd.get_Delta(COSMO, 1.)


def test_get_mass():
    hmd = ccl.halos.MassDef(200, 'critical')
    for R in [1., [1., 2.], np.array([1., 2.])]:
        m = hmd.get_mass(COSMO, R, 1.)
        assert np.all(np.isfinite(m))
        assert np.shape(m) == np.shape(R)


def test_get_radius():
    hmd = ccl.halos.MassDef(200, 'critical')
    for M in [1E12, [1E12, 2E12],
              np.array([1E12, 2E12])]:
        r = hmd.get_radius(COSMO, M, 1.)
        assert np.all(np.isfinite(r))
        assert np.shape(r) == np.shape(M)


def test_translate_mass():
    hmd = ccl.halos.MassDef200m
    hmdb = ccl.halos.MassDef200c
    cm = ccl.halos.Concentration.create_instance("Duffy08", mass_def=hmd)
    translator = ccl.halos.mass_translator(mass_in=hmd, mass_out=hmdb,
                                           concentration=cm)
    for M in [1E12, [1E12, 2E12],
              np.array([1E12, 2E12])]:
        m = translator(COSMO, M, 1)
        assert np.all(np.isfinite(m))
        assert np.shape(m) == np.shape(M)


@pytest.mark.parametrize('scls', [ccl.halos.MassDef200m,
                                  ccl.halos.MassDef200c,
                                  ccl.halos.MassDef500c,
                                  ccl.halos.MassDefVir])
def test_subclasses_smoke(scls):
    hmd = scls
    assert np.isfinite(hmd.get_Delta(COSMO, 1.))


@pytest.mark.parametrize('name', ['200m', '200c', '500c', 'vir', '350m'])
def test_massdef_from_string_smoke(name):
    hmd = ccl.halos.MassDef.from_name(name)
    assert np.isfinite(hmd.get_radius(COSMO, 1e14, 1))


def test_massdef_from_string_raises():
    with pytest.raises(ValueError):
        ccl.halos.MassDef.from_name("my_mass_def")


def test_mass_translator():
    # Check that the mass translator complains for inconsistent masses.
    cm = ccl.halos.Concentration.create_instance("Duffy08")
    mdef1 = ccl.halos.MassDef.create_instance("250c")
    mdef2 = ccl.halos.MassDef.create_instance("500c")
    with pytest.raises(ValueError):
        ccl.halos.mass_translator(mass_in=mdef1, mass_out=mdef2,
                                  concentration=cm)

    # Check that if we pass the same mass definition, it returns the same M.
    mdef1 = mdef2 = cm.mass_def
    translator = ccl.halos.mass_translator(mass_in=mdef1, mass_out=mdef2,
                                           concentration=cm)
    cosmo = ccl.CosmologyVanillaLCDM()
    assert translator(cosmo, 1e14, 1) == 1e14



import numpy as np
import pytest
import pyccl as ccl


NCHI = 100
CHIMIN = 100.
CHIMAX = 1000.
COSMO = ccl.CosmologyVanillaLCDM(transfer_function='bbks')


def tkkaf(k1, k2, a, alpha=1., beta=1.):
    return 1./(k1**alpha*k2**beta)


def get_tk3d(alpha=1, beta=1):
    a_arr = np.linspace(0.1, 1., 10)
    k_arr = np.geomspace(1E-4, 1E3, 10)
    tkka_arr = np.array([tkkaf(k_arr[None, :],
                               k_arr[:, None], a,
                               alpha, beta)
                         for a in a_arr])
    return ccl.Tk3D(a_arr=a_arr, lk_arr=np.log(k_arr),
                    tkk_arr=np.log(tkka_arr),
                    is_logt=True)


def pred_covar(l1, l2, alpha=1., beta=1.,
               prefac=1./(4*np.pi), chi_power=6):
    lfac = 1/((l1+0.5)**alpha*(l2+0.5)**beta)
    xpn = alpha+beta-chi_power+1
    chifac = (CHIMAX**xpn-CHIMIN**xpn)/xpn
    return lfac*chifac*prefac


def get_tracer():
    chis = np.linspace(CHIMIN, CHIMAX, NCHI)
    ws = np.ones(NCHI)
    t = ccl.Tracer()
    t.add_tracer(COSMO, kernel=(chis, ws))
    return t


@pytest.mark.parametrize("alpha,beta,typ", [(3., 3., 'SSC'),
                                            (3., 2., 'SSC'),
                                            (2., 3., 'SSC'),
                                            (2., 2., 'SSC'),
                                            (3., 3., 'cNG'),
                                            (2., 2., 'cNG'),
                                            (1., 2., 'cNG'),
                                            (2., 1., 'cNG')])
def test_cov_NG_sanity(alpha, beta, typ):
    # Compares covariance against analytical prediction
    tsp = get_tk3d(alpha, beta)
    tr = get_tracer()
    ls = np.array([2., 20., 200.])

    if typ == 'cNG':
        cov_p = pred_covar(ls[None, :], ls[:, None], alpha, beta)

        def cov_f(ll, **kwargs):
            return ccl.angular_cl_cov_cNG(
                COSMO, tr, tr, ell=ll, t_of_kk_a=tsp,
                **kwargs)
    elif typ == 'SSC':
        a_s = np.linspace(0.1, 1., 1024)
        s2b = np.ones_like(a_s)
        cov_p = pred_covar(ls[None, :], ls[:, None], alpha, beta,
                           prefac=1., chi_power=4)

        def cov_f(ll, **kwargs):
            return ccl.angular_cl_cov_SSC(
                COSMO, tr, tr, ell=ll, t_of_kk_a=tsp,
                sigma2_B=(a_s, s2b), **kwargs)

    cov = cov_f(ls)
    assert np.all(np.fabs(cov/cov_p-1).flatten() < 1E-5)

    # Spline integration (fast but inaccurate)
    cov = cov_f(ls, integration_method='spline')
    assert np.all(np.fabs(cov/cov_p-1).flatten() < 4E-2)

    # Different tracers
    cov = cov_f(ls, tracer3=tr, tracer4=tr)
    assert np.all(np.fabs(cov/cov_p-1).flatten() < 1E-5)

    # Different ells
    cov = np.array([cov_f(ls, ell2=l)
                    for l in ls])
    assert np.all(np.fabs(cov/cov_p-1).flatten() < 1E-5)

    # Different ells (transpose)
    cov = np.array([cov_f(l, ell2=ls)
                    for l in ls]).T
    assert np.all(np.fabs(cov/cov_p-1).flatten() < 1E-5)

    # All scalars
    cov = np.array([[cov_f(l1, ell2=l2)
                     for l1 in ls]
                    for l2 in ls])
    assert np.all(np.fabs(cov/cov_p-1).flatten() < 1E-5)


@pytest.mark.parametrize("typ", ['cNG', 'SSC'])
def test_cov_NG_errors(typ):
    if typ == 'cNG':
        cov_f = ccl.angular_cl_cov_cNG
    elif typ == 'SSC':
        cov_f = ccl.angular_cl_cov_SSC

    tsp = get_tk3d(1, 1)
    tr = get_tracer()
    ls = np.array([2., 20., 200.])

    with pytest.raises(ValueError):
        cov_f(COSMO, tr, tr, ell=ls, t_of_kk_a=tsp,
              integration_method='cag_cuad')


def test_Sigma2B():
    # Check projected variance calculation against
    # explicit integration
    from scipy.special import jv
    from scipy.integrate import simpson

    fsky = 0.1
    # Default sampling
    a, s2b_a = ccl.sigma2_B_disc(COSMO, fsky=fsky)
    idx = (a > 0.5) & (a < 1)

    a_use = a[idx]
    # Input sampling
    s2b_b = ccl.sigma2_B_disc(COSMO, a_arr=a_use, fsky=fsky)
    # Scalar input sampling
    s2b_c = np.array([ccl.sigma2_B_disc(COSMO, a_arr=a, fsky=fsky)
                      for a in a_use])

    # Alternative calculation
    chis = ccl.comoving_radial_distance(COSMO, a_use)
    Rs = np.arccos(1-2*fsky)*chis

    def integrand(lk, a, R):
        k = np.exp(lk)
        x = k*R
        w = 2*jv(1, x)/x
        pk = ccl.linear_matter_power(COSMO, k, a)
        return w*w*k*k*pk/(2*np.pi)

    lk_arr = np.log(np.geomspace(1E-4, 1E1, 1024))
    s2b_d = np.array([simpson(integrand(lk_arr, a, R), x=lk_arr)
                      for a, R in zip(a_use, Rs)])

    assert np.all(np.fabs(s2b_b/s2b_a[idx]-1) < 1E-10)
    assert np.all(np.fabs(s2b_c/s2b_a[idx]-1) < 1E-10)
    assert np.all(np.fabs(s2b_d/s2b_a[idx]-1) < 1E-3)

    # Check calculation based on mask Cls against the disc calculation
    ell = np.arange(1000)
    theta = np.arccos(1-2*fsky)
    kR = (ell+0.5)*theta
    mask_wl = (ell+0.5)/(2*np.pi) * (2*jv(1, kR)/(kR))**2

    a_use = np.array([0.2, 0.5, 1.0])
    s2b_e = ccl.sigma2_B_from_mask(COSMO, a_arr=a_use, mask_wl=mask_wl)
    s2b_f = ccl.sigma2_B_disc(COSMO, a_arr=a_use, fsky=fsky)
    assert np.all(np.fabs(s2b_e/s2b_f-1) < 1E-3)

    # Test passing a_arr=None (smoke)
    a_s, s2b = ccl.sigma2_B_from_mask(COSMO, a_arr=None,
                                      mask_wl=mask_wl)
    a_cosmo = COSMO.get_pk_spline_a()
    assert np.all(a_s == a_cosmo)
    assert len(a_s) == len(s2b)



import numpy as np
import pyccl as ccl
import pytest


NZ = 128
ZZ = np.linspace(0., 1., NZ)
BZ_C = 2.
BZ = BZ_C * np.ones(NZ)


def test_pt_tracer_m_smoke():
    ccl.nl_pt.PTMatterTracer()


@pytest.mark.parametrize('b2', [(ZZ, BZ), BZ_C, None])
def test_pt_tracer_nc_smoke(b2):
    pt_tr = ccl.nl_pt.PTNumberCountsTracer((ZZ, BZ),
                                           b2=b2,
                                           bs=(ZZ, BZ))

    # Test b1 and bs do the right thing
    for b in [pt_tr.b1, pt_tr.bs]:
        assert b(0.2) == BZ_C

    # Test b2 does the right thing
    if b2 is not None:
        assert pt_tr.b2(0.2) == BZ_C
        zz = np.array([0.2])
        assert pt_tr.b2(zz).squeeze() == BZ_C


@pytest.mark.parametrize('c2', [(ZZ, BZ), BZ_C, None])
def test_pt_tracer_ia_smoke(c2):
    pt_tr = ccl.nl_pt.PTIntrinsicAlignmentTracer((ZZ, BZ),
                                                 c2=c2,
                                                 cdelta=(ZZ, BZ))

    # Test c1 and cdelta do the right thing
    for b in [pt_tr.c1, pt_tr.cdelta]:
        assert b(0.2) == BZ_C

    # Test c2 does the right thing
    if c2 is not None:
        assert pt_tr.c2(0.2) == BZ_C
        zz = np.array([0.2])
        assert pt_tr.c2(zz).squeeze() == BZ_C


def test_pt_tracer_get_bias():
    pt_tr = ccl.nl_pt.PTNumberCountsTracer((ZZ, BZ),
                                           b2=(ZZ, BZ),
                                           bs=(ZZ, BZ))
    b = pt_tr.get_bias('b1', 0.1)
    assert b == BZ_C

    with pytest.raises(KeyError):
        pt_tr.get_bias('b_one', 0.1)



from colossus.halo import mass_so, concentration, mass_defs
from colossus.cosmology import cosmology
import numpy as np

h=0.7
params = {'flat': True, 'H0': 100*h , 'Om0': 0.3, 'Ob0': 0.05, 'sigma8': 0.8, 'ns': 0.96}
cosmo = cosmology.setCosmology('myCosmo', params)

Ms = np.array([1E12, 1E13, 1E14])
Rs_200m = mass_so.M_to_R(Ms * h, 0.0, '200m')/h*0.001
Rs_500c = mass_so.M_to_R(Ms * h, 0.0, '500c')/h*0.001
head_c = 'Mass '
cs_200m_d = concentration.concentration(Ms * h, '200m',
                                        0.0, model = 'duffy08')
head_c += 'c(duffy_200m)'
cs_200c_d = concentration.concentration(Ms * h, '200c',
                                        0.0, model = 'duffy08')
head_c += 'c(duffy_200c)'
cs_200m_b = concentration.concentration(Ms * h, '200m',
                                        0.0, model = 'bhattacharya13')
head_c += 'c(bhattacharya_200m)'
cs_200c_b = concentration.concentration(Ms * h, '200c',
                                        0.0, model = 'bhattacharya13')
head_c += 'c(bhattacharya_200c)'
cs_vir_k = concentration.concentration(Ms * h, 'vir',
                                       0.0, model = 'klypin11')
head_c += 'c(klypin_vir)'
cs_vir_b = concentration.concentration(Ms * h, 'vir',
                                       0.0, model = 'bhattacharya13')
head_c += 'c(bhattacharya_vir)'
cs_200c_p = concentration.concentration(Ms * h, '200c',
                                        0.0, model = 'prada12')
head_c += 'c(prada_200c)'
cs_200c_di = concentration.concentration(Ms * h, '200c',
                                         0.0, model = 'diemer15')
head_c += 'c(diemer_200c)'
Ms_500c, _, _ = mass_defs.changeMassDefinition(Ms * h, cs_200m_d,
                                               0., '200m', '500c')
Ms_500c /= h

np.savetxt("../mdef_bm.txt",
           np.transpose([Ms, Rs_200m, Rs_500c, Ms_500c]),
           header= 'Mass-(200m) Radius-(200m) Radius-(500c), Mass-(500c)')
np.savetxt("../conc_bm.txt",
           np.transpose([Ms,
                         cs_200m_d, cs_200c_d,
                         cs_200m_b, cs_200c_b,
                         cs_vir_k, cs_vir_b,
                         cs_200c_p, cs_200c_di]),
           header = head_c)



__all__ = ("CosmologyParams",)

from ... import lib
from . import CCLParameters


class CosmologyParams(CCLParameters, factory=lib.parameters):
    """Instances of this class hold cosmological parameters."""

    def __getattribute__(self, key):
        if key == "m_nu":
            N_nu_mass = self._instance.N_nu_mass
            nu_masses = lib.parameters_get_nu_masses(self._instance, N_nu_mass)
            return nu_masses.tolist()
        return super().__getattribute__(key)

    def __setattr__(self, key, value):
        if key == "m_nu":
            return lib.parameters_m_nu_set_custom(self._instance, value)
        super().__setattr__(key, value)



import numpy as np
import pyccl as ccl
from velocileptors.EPT.cleft_kexpanded_resummed_fftw import RKECLEFT

cosmo = ccl.Cosmology(Omega_c=0.25, Omega_b=0.05, h=0.7, n_s=0.96, sigma8=0.8,
                      transfer_function='boltzmann_camb', T_CMB=2.725)
lkmin = -4
lkmax = 2
n_per_decade = 20
z_s = np.array([0., 1.])
a_s = 1./(1 + z_s)
ks = np.logspace(lkmin, lkmax, (lkmax - lkmin) * n_per_decade)

pklz0 = cosmo.linear_matter_power(ks, 1.0)
g = cosmo.growth_factor(a_s)

h = cosmo['h']
cleft = RKECLEFT(ks / h, pklz0 * h ** 3)
lpt_table = []
for gz in g:
       cleft.make_ptable(D=gz, kmin=ks[0] / h,
                         kmax=ks[-1] / h, nk=ks.size)
       lpt_table.append(cleft.pktable)
lpt_table = np.array(lpt_table)
lpt_table[:, :, 1:] /= h ** 3

b11 = b12 = 1.3
b21 = b22 = 1.5
bs1 = bs2 = 1.7
b3nl1 = b3nl2 = 1.9
bk21 = bk22 = 0.1

# Transform from Eulerian to Lagrangian biases
bL11 = b11 - 1
bL12 = b12 - 1

# Pgg
Pdmdm = lpt_table[:, :, 1]
Pdmd1 = 0.5*lpt_table[:, :, 2]
Pd1d1 = lpt_table[:, :, 3]
pgg = (Pdmdm + (bL11+bL12) * Pdmd1 +
       (bL11*bL12) * Pd1d1)

Pdmd2 = 0.5*lpt_table[:, :, 4]
Pd1d2 = 0.5*lpt_table[:, :, 5]
Pd2d2 = lpt_table[:, :, 6]
Pdms2 = 0.25*lpt_table[:, :, 7]
Pd1s2 = 0.25*lpt_table[:, :, 8]
Pd2s2 = 0.25*lpt_table[:, :, 9]
Ps2s2 = 0.25*lpt_table[:, :, 10]
Pdmo3 = 0.25*lpt_table[:, :, 11]
Pd1o3 = 0.25*lpt_table[:, :, 12]

Pdmk2 = 0.5*Pdmdm * (ks**2)[None, :]
Pd1k2 = 0.5*Pdmd1 * (ks**2)[None, :]
Pd2k2 = Pdmd2 * (ks**2)[None, :]
Ps2k2 = Pdms2 * (ks**2)[None, :]
Pk2k2 = 0.25*Pdmdm * (ks**4)[None, :]

pgg += ((b21 + b22) * Pdmd2 +
       (bs1 + bs2) * Pdms2 +
       (bL11*b22 + bL12*b21) * Pd1d2 +
       (bL11*bs2 + bL12*bs1) * Pd1s2 +
       (b21*b22) * Pd2d2 +
       (b21*bs2 + b22*bs1) * Pd2s2 +
       (bs1*bs2) * Ps2s2 +
       (b3nl1 + b3nl2) * Pdmo3 +
       (bL11*b3nl2 + bL12*b3nl1) * Pd1o3)

pgg += ((bk21 + bk22) * Pdmk2 +
       (bL12 * bk21 + bL11 * bk22) * Pd1k2 +
       (b22 * bk21 + b21 * bk22) * Pd2k2 +
       (bs2 * bk21 + bs1 * bk22) * Ps2k2 +
       (bk21 * bk22) * Pk2k2)

# Pgm
pgm = Pdmdm + bL11 * Pdmd1
pgm += (b21 * Pdmd2 +
       bs1 * Pdms2 +
       b3nl1 * Pdmo3 +
       bk21 * Pdmk2)

np.savetxt("../pt_bm_lpt_z0.txt",
           np.transpose([ks, pgg[0], pgm[0]]),
           header='[0]-k  [1]-GG [2]-GM')
np.savetxt("../pt_bm_lpt_z1.txt",
           np.transpose([ks, pgg[1], pgm[1]]),
           header='[0]-k  [1]-GG [2]-GM')



import os
import numpy as np
import pyccl as ccl
import pyccl.nl_pt as pt
import pytest

LPTPK_TOLERANCE = 1e-4

# Set cosmology
COSMO = ccl.Cosmology(Omega_c=0.25, Omega_b=0.05,
                      h=0.7, sigma8=0.8, n_s=0.96,
                      T_CMB=2.725,
                      transfer_function='boltzmann_camb')

# Redshifts
zs = np.array([0., 1.])

# Tracers
ptt = {}
ptt['g'] = pt.PTNumberCountsTracer(b1=1.3, b2=1.5, bs=1.7, b3nl=1.9, bk2=0.1)
ptt['m'] = pt.PTMatterTracer()

# Calculator
a_arr = 1./(1+np.array([0., 0.25, 0.5, 0.75, 1.]))[::-1]
ptc = pt.LagrangianPTCalculator(log10k_min=-4, log10k_max=2,
                                nk_per_decade=20, a_arr=a_arr,
                                b1_pk_kind='pt', bk2_pk_kind='pt')
ptc.update_ingredients(COSMO)

# Read data
dirdat = os.path.join(os.path.dirname(__file__), 'data')
data = []
data.append(np.loadtxt(os.path.join(dirdat, 'pt_bm_lpt_z0.txt'), unpack=True))
data.append(np.loadtxt(os.path.join(dirdat, 'pt_bm_lpt_z1.txt'), unpack=True))
order = ['gg', 'gm']


@pytest.mark.parametrize('comb', enumerate(order))
def test_pt_pk(comb):
    i_d, cc = comb
    t1, t2 = cc

    ptt1 = ptt[t1]
    ptt2 = ptt[t2]

    pk = ptc.get_biased_pk2d(ptt1, tracer2=ptt2)

    for iz, z in enumerate(zs):
        a = 1./(1+z)
        kin = data[iz][0]
        ind = np.where((kin < 1.0) & (kin > 1E-3))
        k = kin[ind]
        dpk = data[iz][i_d+1][ind]
        tpk = pk(k, a, COSMO)
        assert np.all(
            np.fabs(tpk / dpk - 1) < LPTPK_TOLERANCE
        )



import tempfile
import filecmp
import io
import pyccl as ccl


def test_yaml():
    cosmo = ccl.Cosmology(Omega_c=0.25, Omega_b=0.05, h=0.7, A_s=2.1e-9,
                          n_s=0.97, m_nu=[0.01, 0.2, 0.3],
                          transfer_function="boltzmann_camb")

    # Make temporary files
    with tempfile.NamedTemporaryFile(delete=True) as tmpfile1, \
            tempfile.NamedTemporaryFile(delete=True) as tmpfile2:
        cosmo.write_yaml(tmpfile1.name)

        cosmo2 = ccl.Cosmology.read_yaml(tmpfile1.name)
        cosmo2.write_yaml(tmpfile2.name)

        # Compare the contents of the two files
        assert filecmp.cmp(tmpfile1.name, tmpfile2.name, shallow=False)

    cosmo = ccl.Cosmology(Omega_c=0.25, Omega_b=0.05, h=0.7, A_s=2.1e-9,
                          n_s=0.97, m_nu=0.1, mass_split="equal",
                          transfer_function="boltzmann_camb")

    stream = io.StringIO()
    cosmo.write_yaml(stream)
    stream.seek(0)

    cosmo2 = ccl.Cosmology.read_yaml(stream)
    stream2 = io.StringIO()
    cosmo2.write_yaml(stream2)

    assert stream.getvalue() == stream2.getvalue()



__all__ = ("get_camb_pk_lin", "get_isitgr_pk_lin", "get_class_pk_lin",)

import numpy as np

try:
    import isitgr  # noqa: F401
except ModuleNotFoundError:
    pass  # prevent nans from isitgr

from . import CCLError, Pk2D, check, lib, sigma8


def get_camb_pk_lin(cosmo, *, nonlin=False):
    """Run CAMB and return the linear power spectrum.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmological
            parameters. The cosmological parameters with
            which to run CAMB.
        nonlin (:obj:`bool`): Whether to compute and return the
            non-linear power spectrum as well.

    Returns:
        :class:`~pyccl.pk2d.Pk2D`: Power spectrum object. The linear power \
            spectrum. If ``nonlin=True``, returns a tuple \
            ``(pk_lin, pk_nonlin)``.
    """
    import camb
    import camb.model

    # Get extra CAMB parameters that were specified
    extra_camb_params = {}
    try:
        extra_camb_params = cosmo["extra_parameters"]["camb"]
    except (KeyError, TypeError):
        pass

    # z sampling from CCL parameters
    na = lib.get_pk_spline_na(cosmo.cosmo)
    status = 0
    a_arr, status = lib.get_pk_spline_a(cosmo.cosmo, na, status)
    check(status, cosmo=cosmo)
    a_arr = np.sort(a_arr)
    zs = 1.0 / a_arr - 1
    zs = np.clip(zs, 0, np.inf)

    # deal with normalization
    if np.isfinite(cosmo["A_s"]):
        A_s_fid = cosmo["A_s"]
    elif np.isfinite(cosmo["sigma8"]):
        A_s_fid = 2.1e-9
        sigma8_target = cosmo["sigma8"]
    else:
        raise CCLError(
            "Could not normalize the linear power spectrum. "
            "A_s = %f, sigma8 = %f" % (
                cosmo['A_s'], cosmo['sigma8']))

    # init camb params
    cp = camb.model.CAMBparams()

    # turn some stuff off
    cp.WantCls = False
    cp.DoLensing = False
    cp.Want_CMB = False
    cp.Want_CMB_lensing = False
    cp.Want_cl_2D_array = False
    cp.WantTransfer = True

    # basic background stuff
    h2 = cosmo['h']**2
    cp.H0 = cosmo['h'] * 100
    cp.ombh2 = cosmo['Omega_b'] * h2
    cp.omch2 = cosmo['Omega_c'] * h2
    cp.omk = cosmo['Omega_k']

    # "constants"
    cp.TCMB = cosmo['T_CMB']

    # neutrinos
    # We maually setup the CAMB neutrinos to match the adjustments CLASS
    # makes to their temperatures.
    cp.share_delta_neff = False
    cp.omnuh2 = cosmo['Omega_nu_mass'] * h2
    cp.num_nu_massless = cosmo['N_nu_rel']
    cp.num_nu_massive = int(cosmo['N_nu_mass'])
    cp.nu_mass_eigenstates = int(cosmo['N_nu_mass'])

    delta_neff = cosmo['Neff'] - 3.046  # used for BBN YHe comps

    # CAMB defines a neutrino degeneracy factor as T_i = g^(1/4)*T_nu
    # where T_nu is the standard neutrino temperature from first order
    # computations
    # CLASS defines the temperature of each neutrino species to be
    # T_i_eff = T_ncdm * T_cmb where T_ncdm is a fudge factor to get the
    # total mass in terms of eV to match second-order computations of the
    # relationship between m_nu and Omega_nu.
    # We are trying to get both codes to use the same neutrino temperature.
    # thus we set T_i_eff = T_i = g^(1/4) * T_nu and solve for the right
    # value of g for CAMB. We get g = (T_ncdm / (11/4)^(-1/3))^4
    g = np.power(
        cosmo["T_ncdm"] / np.power(11.0/4.0, -1.0/3.0),
        4.0)

    if cosmo['N_nu_mass'] > 0:
        nu_mass_fracs = cosmo['m_nu'][:cosmo['N_nu_mass']]
        nu_mass_fracs = nu_mass_fracs / np.sum(nu_mass_fracs)

        cp.nu_mass_numbers = np.ones(cosmo['N_nu_mass'], dtype=np.int)
        cp.nu_mass_fractions = nu_mass_fracs
        cp.nu_mass_degeneracies = np.ones(int(cosmo['N_nu_mass'])) * g
    else:
        cp.nu_mass_numbers = []
        cp.nu_mass_fractions = []
        cp.nu_mass_degeneracies = []

    # get YHe from BBN
    cp.bbn_predictor = camb.bbn.get_predictor()
    cp.YHe = cp.bbn_predictor.Y_He(
        cp.ombh2 * (camb.constants.COBE_CMBTemp / cp.TCMB) ** 3,
        delta_neff)

    camb_de_models = ['DarkEnergyPPF', 'ppf', 'DarkEnergyFluid', 'fluid']
    camb_de_model = extra_camb_params.get('dark_energy_model', 'fluid')
    if camb_de_model not in camb_de_models:
        raise ValueError("The only dark energy models CCL supports with "
                         "CAMB are fluid and ppf.")
    cp.set_classes(
        dark_energy_model=camb_de_model
    )

    if camb_de_model not in camb_de_models[:2] and cosmo['wa'] and \
            (cosmo['w0'] < -1 - 1e-6 or
                1 + cosmo['w0'] + cosmo['wa'] < - 1e-6):
        raise ValueError("If you want to use w crossing -1,"
                         " then please set the dark_energy_model to ppf.")
    cp.DarkEnergy.set_params(
        w=cosmo['w0'],
        wa=cosmo['wa']
    )

    cp.set_for_lmax(extra_camb_params.get("lmax", 5000))
    cp.InitPower.set_params(
        As=A_s_fid,
        ns=cosmo['n_s'])

    cp.set_matter_power(
        redshifts=[_z for _z in zs],
        kmax=extra_camb_params.get("kmax", 10.0))

    camb_res = camb.get_transfer_functions(cp)

    def construct_Pk2D(camb_res, nonlin=False):
        k, z, pk = camb_res.get_linear_matter_power_spectrum(
            hubble_units=True, nonlinear=nonlin)

        # convert to non-h inverse units
        k *= cosmo['h']
        pk /= (h2 * cosmo['h'])

        # now build interpolant
        nk = k.shape[0]
        lk_arr = np.log(k)
        a_arr = 1.0 / (1.0 + z)
        na = a_arr.shape[0]
        sinds = np.argsort(a_arr)
        a_arr = a_arr[sinds]
        ln_p_k_and_z = np.zeros((na, nk), dtype=np.float64)
        for i, sind in enumerate(sinds):
            ln_p_k_and_z[i, :] = np.log(pk[sind, :])

        pk = Pk2D(
            a_arr=a_arr,
            lk_arr=lk_arr,
            pk_arr=ln_p_k_and_z,
            is_logp=True,
            extrap_order_lok=1,
            extrap_order_hik=2)

        return pk

    if np.isfinite(cosmo["sigma8"]):
        camb_res.calc_power_spectra()
        pk = construct_Pk2D(camb_res, nonlin=False)
        sigma8_tmp = sigma8(cosmo, p_of_k_a=pk)
        camb_res.Params.InitPower.As *= sigma8_target**2 / sigma8_tmp**2

    if nonlin:
        camb_res.Params.NonLinear = camb.model.NonLinear_pk
        camb_res.Params.NonLinearModel = camb.nonlinear.Halofit()
        halofit_version = extra_camb_params.get("halofit_version", "mead")
        options = {k: extra_camb_params[k] for k in
                   ["HMCode_A_baryon",
                    "HMCode_eta_baryon",
                    "HMCode_logT_AGN"] if k in extra_camb_params}
        camb_res.Params.NonLinearModel.set_params(
            halofit_version=halofit_version,
            **options)
    else:
        assert camb_res.Params.NonLinear == camb.model.NonLinear_none

    camb_res.calc_power_spectra()
    pk_lin = construct_Pk2D(camb_res, nonlin=False)

    if not nonlin:
        return pk_lin
    else:
        pk_nonlin = construct_Pk2D(camb_res, nonlin=True)

        return pk_lin, pk_nonlin


def get_isitgr_pk_lin(cosmo):
    """Run ISiTGR-CAMB and return the linear power spectrum.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmological
            parameters. The cosmological parameters with
            which to run ISiTGR-CAMB.

    Returns:
        :class:`~pyccl.pk2d.Pk2D`: Power spectrum \
            object. The linear power spectrum.
    """
    import isitgr  # noqa: F811
    import isitgr.model

    # Get extra CAMB parameters that were specified
    extra_camb_params = {}
    try:
        extra_camb_params = cosmo["extra_parameters"]["camb"]
    except (KeyError, TypeError):
        pass

    # z sampling from CCL parameters
    na = lib.get_pk_spline_na(cosmo.cosmo)
    status = 0
    a_arr, status = lib.get_pk_spline_a(cosmo.cosmo, na, status)
    check(status, cosmo=cosmo)
    a_arr = np.sort(a_arr)
    zs = 1.0 / a_arr - 1
    zs = np.clip(zs, 0, np.inf)

    # deal with normalization
    if np.isfinite(cosmo["A_s"]):
        A_s_fid = cosmo["A_s"]
    elif np.isfinite(cosmo["sigma8"]):
        # in this case, CCL will internally normalize for us when we init
        # the linear power spectrum - so we just get close
        A_s_fid = 2.43e-9 * (cosmo["sigma8"] / 0.87659)**2
    else:
        raise CCLError(
            "Could not normalize the linear power spectrum. "
            "A_s = %f, sigma8 = %f" % (
                cosmo['A_s'], cosmo['sigma8']))

    # init isitgr params
    cp = isitgr.model.CAMBparams()

    # turn some stuff off
    cp.WantCls = False
    cp.DoLensing = False
    cp.Want_CMB = False
    cp.Want_CMB_lensing = False
    cp.Want_cl_2D_array = False
    cp.WantTransfer = True

    # basic background stuff
    h2 = cosmo['h']**2
    cp.H0 = cosmo['h'] * 100
    cp.ombh2 = cosmo['Omega_b'] * h2
    cp.omch2 = cosmo['Omega_c'] * h2
    cp.omk = cosmo['Omega_k']
    cp.GR = 1  # means GR modified!
    cp.ISiTGR_muSigma = True
    cp.mu0 = cosmo.mg_parametrization.mu_0
    cp.Sigma0 = cosmo.mg_parametrization.sigma_0
    cp.c1 = cosmo.mg_parametrization.c1_mg
    cp.c2 = cosmo.mg_parametrization.c2_mg
    cp.Lambda = cosmo.mg_parametrization.lambda_mg

    # "constants"
    cp.TCMB = cosmo['T_CMB']

    # neutrinos
    # We maually setup the CAMB neutrinos to match the adjustments CLASS
    # makes to their temperatures.
    cp.share_delta_neff = False
    cp.omnuh2 = cosmo['Omega_nu_mass'] * h2
    cp.num_nu_massless = cosmo['N_nu_rel']
    cp.num_nu_massive = int(cosmo['N_nu_mass'])
    cp.nu_mass_eigenstates = int(cosmo['N_nu_mass'])

    delta_neff = cosmo['Neff'] - 3.046  # used for BBN YHe comps

    # ISiTGR built on CAMB which defines a neutrino degeneracy
    # factor as T_i = g^(1/4)*T_nu
    # where T_nu is the standard neutrino temperature from first order
    # computations
    # CLASS defines the temperature of each neutrino species to be
    # T_i_eff = T_ncdm * T_cmb where T_ncdm is a fudge factor to get the
    # total mass in terms of eV to match second-order computations of the
    # relationship between m_nu and Omega_nu.
    # We are trying to get both codes to use the same neutrino temperature.
    # thus we set T_i_eff = T_i = g^(1/4) * T_nu and solve for the right
    # value of g for CAMB. We get g = (T_ncdm / (11/4)^(-1/3))^4
    g = np.power(
        cosmo["T_ncdm"] / np.power(11.0/4.0, -1.0/3.0),
        4.0)

    if cosmo['N_nu_mass'] > 0:
        nu_mass_fracs = cosmo['m_nu'][:cosmo['N_nu_mass']]
        nu_mass_fracs = nu_mass_fracs / np.sum(nu_mass_fracs)

        cp.nu_mass_numbers = np.ones(cosmo['N_nu_mass'], dtype=np.int)
        cp.nu_mass_fractions = nu_mass_fracs
        cp.nu_mass_degeneracies = np.ones(int(cosmo['N_nu_mass'])) * g
    else:
        cp.nu_mass_numbers = []
        cp.nu_mass_fractions = []
        cp.nu_mass_degeneracies = []

    # get YHe from BBN
    cp.bbn_predictor = isitgr.bbn.get_predictor()
    cp.YHe = cp.bbn_predictor.Y_He(
        cp.ombh2 * (isitgr.constants.COBE_CMBTemp / cp.TCMB) ** 3,
        delta_neff)

    camb_de_models = ['DarkEnergyPPF', 'ppf', 'DarkEnergyFluid', 'fluid']
    camb_de_model = extra_camb_params.get('dark_energy_model', 'fluid')
    if camb_de_model not in camb_de_models:
        raise ValueError("The only dark energy models CCL supports with "
                         "CAMB are fluid and ppf.")
    cp.set_classes(
        dark_energy_model=camb_de_model
    )
    if camb_de_model not in camb_de_models[:2] and cosmo['wa'] and \
            (cosmo['w0'] < -1 - 1e-6 or
                1 + cosmo['w0'] + cosmo['wa'] < - 1e-6):
        raise ValueError("For w to cross -1, set `dark_energy_model=ppf`.")
    cp.DarkEnergy.set_params(
        w=cosmo['w0'],
        wa=cosmo['wa']
    )
    # cp.set_cosmology()
    cp.set_matter_power(
        redshifts=[_z for _z in zs],
        kmax=10,
        nonlinear=False)
    assert cp.NonLinear == isitgr.model.NonLinear_none

    cp.set_for_lmax(5000)
    cp.InitPower.set_params(
        As=A_s_fid,
        ns=cosmo['n_s'])

    # run ISITGR and get results
    isitgr_res = isitgr.get_results(cp)
    k, z, pk = isitgr_res.get_linear_matter_power_spectrum(
        hubble_units=True, nonlinear=False)

    # convert to non-h inverse units
    k *= cosmo['h']
    pk /= (h2 * cosmo['h'])

    # now build interpolant
    nk = k.shape[0]
    lk_arr = np.log(k)
    a_arr = 1.0 / (1.0 + z)
    na = a_arr.shape[0]
    sinds = np.argsort(a_arr)
    a_arr = a_arr[sinds]
    ln_p_k_and_z = np.zeros((na, nk), dtype=np.float64)
    for i, sind in enumerate(sinds):
        ln_p_k_and_z[i, :] = np.log(pk[sind, :])

    pk_lin = Pk2D(
        a_arr=a_arr,
        lk_arr=lk_arr,
        pk_arr=ln_p_k_and_z,
        is_logp=True,
        extrap_order_lok=1,
        extrap_order_hik=2)
    return pk_lin


def get_class_pk_lin(cosmo):
    """Run CLASS and return the linear power spectrum.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmological
            parameters. The cosmological parameters with
            which to run CLASS.

    Returns:
        :class:`~pyccl.pk2d.Pk2D`: Power spectrum object.\
            The linear power spectrum.
    """
    import classy

    params = {
        "output": "mPk",
        "non linear": "none",
        "P_k_max_1/Mpc": cosmo.cosmo.spline_params.K_MAX_SPLINE,
        "z_max_pk": 1.0/cosmo.cosmo.spline_params.A_SPLINE_MINLOG_PK-1.0,
        "modes": "s",
        "lensing": "no",
        "h": cosmo["h"],
        "Omega_cdm": cosmo["Omega_c"],
        "Omega_b": cosmo["Omega_b"],
        "Omega_k": cosmo["Omega_k"],
        "n_s": cosmo["n_s"]}

    # cosmological constant?
    # set Omega_Lambda = 0.0 if w !=-1 or wa != 0
    if cosmo['w0'] != -1 or cosmo['wa'] != 0:
        params["Omega_Lambda"] = 0
        params['w0_fld'] = cosmo['w0']
        params['wa_fld'] = cosmo['wa']

    # neutrino parameters
    # massless neutrinos
    if cosmo["N_nu_rel"] > 1e-4:
        params["N_ur"] = cosmo["N_nu_rel"]
    else:
        params["N_ur"] = 0.0

    # massive neutrinos
    if cosmo["N_nu_mass"] > 0:
        params["N_ncdm"] = cosmo["N_nu_mass"]
        masses = lib.parameters_get_nu_masses(cosmo._params, 3)
        params["m_ncdm"] = ", ".join(
            ["%g" % m for m in masses[:cosmo["N_nu_mass"]]])

    params["T_cmb"] = cosmo["T_CMB"]

    # if we have sigma8, we need to find A_s
    if np.isfinite(cosmo["A_s"]):
        params["A_s"] = cosmo["A_s"]
    elif np.isfinite(cosmo["sigma8"]):
        # in this case, CCL will internally normalize for us when we init
        # the linear power spectrum - so we just get close
        A_s_fid = 2.43e-9 * (cosmo["sigma8"] / 0.87659)**2
        params["A_s"] = A_s_fid
    else:
        raise CCLError(
            "Could not normalize the linear power spectrum. "
            "A_s = %f, sigma8 = %f" % (
                cosmo['A_s'], cosmo['sigma8']))

    model = None
    try:
        model = classy.Class()
        model.set(params)
        model.compute()

        # Set k and a sampling from CCL parameters
        nk = lib.get_pk_spline_nk(cosmo.cosmo)
        na = lib.get_pk_spline_na(cosmo.cosmo)
        status = 0
        a_arr, status = lib.get_pk_spline_a(cosmo.cosmo, na, status)
        check(status, cosmo=cosmo)

        # FIXME - getting the lowest CLASS k value from the python interface
        # appears to be broken - setting to 1e-5 which is close to the
        # old value
        lk_arr = np.log(np.logspace(
            -5,
            np.log10(cosmo.cosmo.spline_params.K_MAX_SPLINE), nk))

        # we need to cut this to the max value used for calling CLASS
        msk = lk_arr < np.log(cosmo.cosmo.spline_params.K_MAX_SPLINE)
        nk = int(np.sum(msk))
        lk_arr = lk_arr[msk]

        # now do interp by hand
        ln_p_k_and_z = np.zeros((na, nk), dtype=np.float64)
        for aind in range(na):
            z = max(1.0 / a_arr[aind] - 1, 1e-10)
            for kind in range(nk):
                ln_p_k_and_z[aind, kind] = np.log(
                    model.pk_lin(np.exp(lk_arr[kind]), z))
    finally:
        if model is not None:
            model.struct_cleanup()
            model.empty()

    params["P_k_max_1/Mpc"] = cosmo.cosmo.spline_params.K_MAX_SPLINE

    # make the Pk2D object
    pk_lin = Pk2D(
        a_arr=a_arr,
        lk_arr=lk_arr,
        pk_arr=ln_p_k_and_z,
        is_logp=True,
        extrap_order_lok=1,
        extrap_order_hik=2)

    return pk_lin



import numpy as np
import pytest
import pyccl as ccl


COSMO = ccl.Cosmology(
    Omega_c=0.27, Omega_b=0.05, h=0.67, sigma8=0.8, n_s=0.96,
    transfer_function='bbks', matter_power_spectrum='linear')
HMFS = [ccl.halos.MassFuncPress74,
        ccl.halos.MassFuncSheth99,
        ccl.halos.MassFuncJenkins01,
        ccl.halos.MassFuncAngulo12,
        ccl.halos.MassFuncTinker08,
        ccl.halos.MassFuncTinker10,
        ccl.halos.MassFuncWatson13,
        ccl.halos.MassFuncDespali16,
        ccl.halos.MassFuncBocquet16,
        ccl.halos.MassFuncBocquet20]
MS = [1E13, [1E12, 1E15], np.array([1E12, 1E15])]
MFOF = ccl.halos.MassDef('fof', 'matter')
MVIR = ccl.halos.MassDef('vir', 'critical')
M100 = ccl.halos.MassDef(100, 'matter')
M200c = ccl.halos.MassDef(200, 'critical')
M200m = ccl.halos.MassDef(200, 'matter')
M500c = ccl.halos.MassDef(500, 'critical')
M500m = ccl.halos.MassDef(500, 'matter')
MDFS = [MVIR, MVIR, MVIR, MVIR,
        MFOF, MFOF, MVIR, MFOF, MFOF, MFOF]
# This is kind of slow to initialize, so let's do it only once
MF_emu = ccl.halos.MassFuncBocquet20(mass_def='200c')


@pytest.mark.parametrize('nM_class', HMFS)
def test_nM_subclasses_smoke(nM_class):
    nM = nM_class()
    for m in MS:
        n = nM(COSMO, m, 0.9)
        assert np.all(np.isfinite(n))
        assert np.shape(n) == np.shape(m)


@pytest.mark.parametrize('nM_pair', zip(HMFS, MDFS))
def test_nM_mdef_raises(nM_pair):
    nM_class, mdef = nM_pair
    with pytest.raises(ValueError):
        nM_class(mass_def=mdef)


@pytest.mark.parametrize('nM_class', [ccl.halos.MassFuncTinker08,
                                      ccl.halos.MassFuncTinker10])
def test_nM_mdef_bad_delta(nM_class):
    with pytest.raises(ValueError):
        nM_class(mass_def=MFOF)


@pytest.mark.parametrize('nM_class', [ccl.halos.MassFuncTinker08,
                                      ccl.halos.MassFuncTinker10])
def test_nM_SO_allgood(nM_class):
    nM = nM_class(mass_def=MVIR)
    for m in MS:
        n = nM(COSMO, m, 0.9)
        assert np.all(np.isfinite(n))
        assert np.shape(n) == np.shape(m)


def test_nM_despali_smoke():
    nM = ccl.halos.MassFuncDespali16(ellipsoidal=True)
    for m in MS:
        n = nM(COSMO, m, 0.9)
        assert np.all(np.isfinite(n))
        assert np.shape(n) == np.shape(m)


@pytest.mark.parametrize('mdef', [MFOF, M200m])
def test_nM_watson_smoke(mdef):
    nM = ccl.halos.MassFuncWatson13(mass_def=mdef)
    for m in MS:
        n = nM(COSMO, m, 0.9)
        assert np.all(np.isfinite(n))
        assert np.shape(n) == np.shape(m)
    for m in MS:
        n = nM(COSMO, m, 0.1)
        assert np.all(np.isfinite(n))
        assert np.shape(n) == np.shape(m)


@pytest.mark.parametrize('with_hydro', [True, False])
def test_nM_bocquet_smoke(with_hydro):
    with pytest.raises(ValueError):
        ccl.halos.MassFuncBocquet16(mass_def=M500m, hydro=with_hydro)

    for md in [M500c, M200c, M200m]:
        nM = ccl.halos.MassFuncBocquet16(mass_def=md, hydro=with_hydro)
        for m in MS:
            n = nM(COSMO, m, 0.9)
            assert np.all(np.isfinite(n))
            assert np.shape(n) == np.shape(m)


@pytest.mark.parametrize('name', ['Press74', 'Tinker08',
                                  'Despali16', 'Angulo12'])
def test_nM_from_string(name):
    nM_class = ccl.halos.MassFunc.from_name(name)
    nM = nM_class()
    for m in MS:
        n = nM(COSMO, m, 0.9)
        assert np.all(np.isfinite(n))
        assert np.shape(n) == np.shape(m)


def test_nM_from_string_raises():
    with pytest.raises(KeyError):
        ccl.halos.MassFunc.from_name('Tinker09')


@pytest.mark.parametrize('mf', [ccl.halos.MassFuncTinker08,
                                ccl.halos.MassFuncTinker10])
def test_nM_tinker_crit(mf):
    a = 0.5
    om = ccl.omega_x(COSMO, a, 'matter')
    oc = ccl.omega_x(COSMO, a, 'critical')
    delta_c = 500.
    delta_m = delta_c * oc / om
    mdef_c = ccl.halos.MassDef(delta_c, 'critical')
    mdef_m = ccl.halos.MassDef(delta_m, 'matter')
    nM_c = mf(mass_def=mdef_c)
    nM_m = mf(mass_def=mdef_m)
    assert np.allclose(nM_c(COSMO, 1E13, a), nM_m(COSMO, 1E13, a))


def test_nM_tinker10_norm():
    from scipy.integrate import quad

    md = ccl.halos.MassDef(300, rho_type='matter')
    mf = ccl.halos.MassFuncTinker10(mass_def=md, norm_all_z=True)
    bf = ccl.halos.HaloBiasTinker10(mass_def=md)

    def integrand(lnu, z):
        nu = np.exp(lnu)
        a = 1./(1+z)
        gnu = mf._get_fsigma(COSMO, 1.686/nu, a, 1)
        bnu = bf._get_bsigma(COSMO, bf.dc/nu, a)
        return gnu*bnu

    def norm(z):
        return quad(integrand, -13, 2, args=(z,))[0]

    zs = np.linspace(0, 1, 4)
    ns = np.array([norm(z) for z in zs])
    assert np.all(np.fabs(ns-1) < 0.005)


def test_mass_function_mass_def_strict_always_raises():
    # Verify that when the property `_mass_def_strict_always` is set to True,
    # the `mass_def_strict` check cannot be relaxed.
    mdef = ccl.halos.MassDef(400, "critical")
    with pytest.raises(ValueError):
        ccl.halos.MassFuncBocquet16(mass_def=mdef, mass_def_strict=False)


def test_nM_bocquet20_compare():
    # Check that the values are sensible (they don't depart from other
    # parametrisations by more than ~10%
    Ms = np.geomspace(1E14, 1E15, 128)
    mf1 = MF_emu
    mf2 = ccl.halos.MassFuncTinker08(mass_def='200c')

    nM1 = mf1(COSMO, Ms, 1.0)
    nM2 = mf2(COSMO, Ms, 1.0)
    assert np.allclose(nM1, nM2, atol=0, rtol=0.1)


def test_nM_bocquet20_raises():
    Ms = np.geomspace(1E12, 1E17, 128)

    # Need sigma8
    cosmo = ccl.Cosmology(Omega_c=0.25, Omega_b=0.05, h=0.67,
                          A_s=2E-9, n_s=0.96)
    with pytest.raises(ValueError):
        MF_emu(cosmo, Ms, 1.0)

    # Cosmo parameters out of bounds
    cosmo = ccl.Cosmology(Omega_c=0.25, Omega_b=0.05, h=0.67,
                          sigma8=0.8, n_s=2.0)
    with pytest.raises(ValueError):
        MF_emu(cosmo, Ms, 1.0)

    # Redshift out of bounds
    with pytest.raises(ValueError):
        MF_emu(cosmo, Ms, 0.3)



import numpy as np
import pyccl as ccl


def test_hodcl():
    # With many thanks to Ryu Makiya, Eiichiro Komatsu
    # and Shin'ichiro Ando for providing this benchmark.
    # HOD params
    log10Mcut = 11.8
    log10M1 = 11.73
    sigma_Ncen = 0.15
    alp_Nsat = 0.77
    rmax = 4.39
    rgs = 1.17

    # Input power spectrum
    ks = np.loadtxt("benchmarks/data/k_hod.txt")
    zs = np.loadtxt("benchmarks/data/z_hod.txt")
    pks = np.loadtxt("benchmarks/data/pk_hod.txt")
    l_bm, cl_bm = np.loadtxt("benchmarks/data/cl_hod.txt",
                             unpack=True)

    # Set N(z)
    def _nz_2mrs(z):
        # From 1706.05422
        m = 1.31
        beta = 1.64
        x = z / 0.0266
        return x**m * np.exp(-x**beta)
    z1 = 1e-5
    z2 = 0.1
    z_arr = np.linspace(z1, z2, 1024)
    dndz = _nz_2mrs(z_arr)

    # CCL prediction
    # Make sure we use the same P(k)
    cosmo = ccl.CosmologyCalculator(
        Omega_b=0.05,
        Omega_c=0.25,
        h=0.67,
        n_s=0.9645,
        A_s=2.0E-9,
        m_nu=0.00001,
        mass_split='equal',
        pk_linear={'a': 1./(1.+zs[::-1]),
                   'k': ks,
                   'delta_matter:delta_matter': pks[::-1, :]})
    cosmo.compute_growth()

    # Halo model setup
    mass_def = ccl.halos.MassDef(200, 'critical')
    cm = ccl.halos.ConcentrationDuffy08(mass_def=mass_def)
    hmf = ccl.halos.MassFuncTinker08(mass_def=mass_def)
    hbf = ccl.halos.HaloBiasTinker10(mass_def=mass_def)
    hmc = ccl.halos.HMCalculator(mass_function=hmf, halo_bias=hbf,
                                 mass_def=mass_def)
    prf = ccl.halos.HaloProfileHOD(
        mass_def=mass_def,
        concentration=cm,
        log10Mmin_0=np.log10(10.**log10Mcut/cosmo['h']),
        siglnM_0=sigma_Ncen,
        log10M0_0=np.log10(10.**log10Mcut/cosmo['h']),
        log10M1_0=np.log10(10.**log10M1/cosmo['h']),
        alpha_0=alp_Nsat,
        bg_0=rgs,
        bmax_0=rmax)
    prf2pt = ccl.halos.Profile2ptHOD()
    # P(k)
    a_arr, lk_arr, _ = cosmo.get_linear_power().get_spline_arrays()
    pk_hod = ccl.halos.halomod_Pk2D(cosmo, hmc, prf, prof_2pt=prf2pt,
                                    lk_arr=lk_arr, a_arr=a_arr)
    # C_ell
    tr = ccl.NumberCountsTracer(cosmo, has_rsd=False, dndz=(z_arr, dndz),
                                bias=(z_arr, np.ones(len(dndz))))
    cl_hod = ccl.angular_cl(cosmo, tr, tr, ell=l_bm, p_of_k_a=pk_hod)

    assert np.all(np.fabs(cl_hod/cl_bm-1) < 0.005)



import numpy as np
import pyccl as ccl
import pytest

SIGMAM_TOLERANCE = 3.0E-5


@pytest.mark.parametrize(
    'model,w0,wa',
    [(1, -1.0, 0.0),
     (2, -0.9, 0.0),
     (3, -0.9, 0.1)])
def test_sigmaM(model, w0, wa):
    cosmo = ccl.Cosmology(
        Omega_c=0.25,
        Omega_b=0.05,
        h=0.7,
        sigma8=0.8,
        n_s=0.96,
        Neff=0,
        m_nu=0.0,
        w0=w0,
        wa=wa,
        T_CMB=2.7,
        mass_split='normal',
        Omega_g=0,
        Omega_k=0,
        transfer_function='bbks',
        matter_power_spectrum='linear')

    data = np.loadtxt('./benchmarks/data/model%d_sm.txt' % model)

    for i in range(data.shape[0]):
        m = data[i, 0] / cosmo['h']
        sm = ccl.sigmaM(cosmo, m, 1)
        err = sm / data[i, 1] - 1
        np.allclose(err, 0, rtol=0, atol=SIGMAM_TOLERANCE)



import numpy as np
import pyccl as ccl
import matplotlib.pyplot as plt
from scipy.integrate import quad
from scipy.interpolate import interp1d
import os
from scipy.special import sici

cpar1={'Om_m':0.3,'Om_b':0.05,'Om_nu':0.,
       'h':0.70,'sig8':0.8,'n':0.96}
cpar2={'Om_m':0.272,'Om_b':0.0455,'Om_nu':0.,
       'h':0.704,'sig8':0.81,'n':0.967}
cpar3={'Om_m':0.3175,'Om_b':0.0490,'Om_nu':0.,
       'h':0.6711,'sig8':0.834,'n':0.9624}

USE_INT=False

LKHMIN=-4.
LKHMAX=1.
NKH=256

NM=2048
LM0=5.5
LMMIN=6.
LMMAX=16.

params={'DELTA_C':1.686470199841,
        'ST_A': 0.322,
        'ST_a': 0.707,
        'ST_p': 0.3,
        'DUF_A':7.85,
        'DUF_B':-0.081,
        'DUF_C':-0.71}

def st_gs(sigma,dc) :
    nu=dc/sigma
    anu2=params['ST_a']*nu*nu
    expnu=np.exp(-anu2*0.5)

    return params['ST_A']*np.sqrt(2*params['ST_a']/np.pi)*(1+(1./anu2)**params['ST_p'])*nu*expnu

def st_b1(sigma,dc) :
    nu=dc/sigma
    anu2=params['ST_a']*nu*nu

    return 1+(anu2-1+2*params['ST_p']/(1+anu2**params['ST_p']))/params['DELTA_C']

def duffy_c(m,z) :
    return params['DUF_A']*(m/2E12)**params['DUF_B']*(1+z)**params['DUF_C']

def get_halomod_pk(cp,karr,z,integrate=True,fname_out=None) :
    cosmo=ccl.Cosmology(Omega_c=cp['Om_m']-cp['Om_b'],Omega_b=cp['Om_b'],h=cp['h'],
                        sigma8=cp['sig8'],n_s=cp['n'],
                        transfer_function='eisenstein_hu',matter_power_spectrum='linear')

    lmarr=LMMIN+(LMMAX-LMMIN)*(np.arange(NM)+0.5)/NM #log(M*h/M_sum)
    dlm=(LMMAX-LMMIN)/NM
    lmarrb=np.zeros(NM+2); lmarrb[0]=lmarr[0]-dlm; lmarrb[1:-1]=lmarr; lmarrb[-1]=lmarr[-1]+dlm
    marr=10.**lmarr #M (in Msun/h) M_h Ms/h = M Ms
    marrb=10.**lmarrb #M (in Msun/h) M_h Ms/h = M Ms
    sigmarrb=ccl.sigmaM(cosmo,marrb/cp['h'],1./(1+z))
    sigmarr=sigmarrb[1:-1]
    dlsMdl10M=np.fabs((sigmarrb[2:]-sigmarrb[:-2])/(2*dlm))/sigmarr
    omega_z=cp['Om_m']*(1+z)**3/(cp['Om_m']*(1+z)**3+1-cp['Om_m'])
    delta_c=params['DELTA_C']*(1+0.012299*np.log10(omega_z))
    Delta_v=(18*np.pi**2+82*(omega_z-1)-39*(omega_z-1)**2)/omega_z
    fM=st_gs(sigmarr,delta_c)*dlsMdl10M
    bM=st_b1(sigmarr,delta_c)
    cM=duffy_c(marr,z)
    cMf=interp1d(lmarr,cM,kind='linear',bounds_error=False,fill_value=cM[0])
    rhoM=ccl.rho_x(cosmo,1.,'matter')/cp['h']**2

    #Compute mass function normalization
    fM0=1-np.sum(fM)*dlm
    fbM0=1-np.sum(fM*bM)*dlm

    rvM=(3*marr/(4*np.pi*rhoM*Delta_v))**0.333333333333
    rsM=rvM/cM
    rsM0=(3*10.**LM0/(4*np.pi*rhoM*Delta_v))**0.333333333/duffy_c(10.**LM0,z)
    rsMf=interp1d(lmarr,rsM,kind='linear',bounds_error=False,fill_value=rsM0)

    f1hf=interp1d(lmarr,marr*fM/rhoM,kind='linear',bounds_error=False,fill_value=0)
    f1h0=fM0*10.**LM0/rhoM
    f2hf=interp1d(lmarr,fM*bM,kind='linear',bounds_error=False,fill_value=0)
    f2h0=fbM0

    #NFW profile
    def u_nfw(lm,k) :
        x=k*rsMf(lm) #k*r_s
        c=cMf(lm)
        sic,cic=sici(x*(1+c))
        six,cix=sici(x)
        fsin=np.sin(x)*(sic-six)
        fcos=np.cos(x)*(cic-cix)
        fsic=np.sin(c*x)/(x*(1+c))
        fcon=np.log(1.+c)-c/(1+c)

        return (fsin+fcos-fsic)/fcon

    def p1h(k) :
        def integ(lm) :
            u=u_nfw(lm,k)
            f1h=f1hf(lm)
            return u*u*f1h
        if integrate :
            return quad(integ,lmarr[0],lmarr[-1])[0]+f1h0*(u_nfw(LM0,k))**2
        else :
            return np.sum(integ(lmarr))*dlm+f1h0*(u_nfw(LM0,k))**2
    
    def b2h(k) :
        def integ(lm) :
            u=u_nfw(lm,k)
            f2h=f2hf(lm)
            return u*f2h
        if integrate :
            return quad(integ,lmarr[0],lmarr[-1])[0]+f2h0*u_nfw(LM0,k)
        else :
            return np.sum(integ(lmarr))*dlm+f2h0*u_nfw(LM0,k)


    p1harr=np.array([p1h(kk) for kk in karr])
    b2harr=np.array([b2h(kk) for kk in karr])
    pklin=ccl.linear_matter_power(cosmo,karr*cp['h'],1./(1+z))*cp['h']**3

    p2harr=pklin*b2harr**2
    ptarr=p1harr+p2harr
    np.savetxt(fname_out,np.transpose([karr,pklin,p2harr,p1harr,ptarr]))
    return pklin,pklin*b2harr**2,p1harr,pklin*b2harr**2+p1harr

k_arr=10.**(LKHMIN+(LKHMAX-LKHMIN)*(np.arange(NKH)+0.5)/NKH)
for ip,cp in enumerate([cpar1,cpar2,cpar3]) :
    for z in [0,1] :
        get_halomod_pk(cp,k_arr,z+0.,integrate=USE_INT,fname_out='pk_hm_c%d_z%d.txt'%(ip+1,z))



#!/usr/bin/env python
from setuptools.command.build_py import build_py as _build
from setuptools.command.develop import develop as _develop
from distutils.command.clean import clean as _clean
from setuptools import Distribution as _distribution
from setuptools import setup, find_packages
from subprocess import call
from io import open
import os
import shutil
import sys


def _compile_ccl(debug=False):
    call(["mkdir", "-p", "build"])
    v = sys.version_info
    cmd = ["cmake", "-H.", "-Bbuild",
           "-DPYTHON_VERSION=%d.%d.%d" % (v.major, v.minor, v.micro)]
    if debug:
        cmd += ["-DCMAKE_BUILD_TYPE=Debug"]
    if call(cmd) != 0:
        raise Exception(
            "Could not run CMake configuration. Make sure "
            "CMake is installed !")

    if call(["make", "-Cbuild", "_ccllib"]) != 0:
        raise Exception("Could not build CCL")

    # Finds the library under its different possible names
    if os.path.exists("build/pyccl/_ccllib.so"):
        call(["cp", "build/pyccl/_ccllib.so", "pyccl/"])
    else:
        raise Exception("Could not find wrapper shared library, "
                        "compilation must have failed.")
    if call(["cp", "build/pyccl/ccllib.py", "pyccl/"]) != 0:
        raise Exception("Could not find python module, "
                        "SWIG must have failed.")


class Distribution(_distribution):
    global_options = _distribution.global_options

    global_options += [
        ("debug", None, "Debug build"),
    ]

    def __init__(self, attr=None):
        self.debug = False
        super().__init__(attr)


class Build(_build):
    """Specialized Python source builder."""

    def run(self):
        _compile_ccl(debug=self.distribution.debug)
        _build.run(self)


class Develop(_develop):
    """Specialized Python develop mode."""
    def run(self):
        _compile_ccl(debug=self.distribution.debug)
        _develop.run(self)


class Clean(_clean):
    """Remove the copied _ccllib.so"""
    def run(self):
        if os.path.isfile("pyccl/_ccllib.so"):
            os.remove("pyccl/_ccllib.so")
            print("Removed pyccl/_ccllib.so")
        _clean.run(self)
        shutil.rmtree("build")
        print("Removed build.")


# read the contents of the README file
with open('README.md', encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="pyccl",
    description="Library of validated cosmological functions.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="LSST DESC",
    url="https://github.com/LSSTDESC/CCL",
    packages=find_packages(),
    provides=['pyccl'],
    package_data={'pyccl': ['_ccllib.so', 'emulators/data/*.npz']},
    include_package_data=True,
    use_scm_version=True,
    distclass=Distribution,
    python_requires='>=3.8',
    setup_requires=['setuptools_scm'],
    install_requires=['numpy'],
    cmdclass={'build_py': Build, 'develop': Develop, 'clean': Clean},
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: C',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Astronomy',
    ]
)



__all__ = (
    "CCLParameters", "DefaultParams", "SplineParams", "GSLParams",
    "PhysicalConstants", "spline_params", "gsl_params", "physical_constants",
    "DEFAULT_POWER_SPECTRUM",)

from ... import lib

DEFAULT_POWER_SPECTRUM = "delta_matter:delta_matter"


class DefaultParams:
    """Default cosmological parameters used throughout the library.
    """
    #: Mean CMB temperature in Kelvin.
    T_CMB = 2.7255
    #: Non-CDM temperature in units of ``T_CMB``.
    T_ncdm = 0.71611


class CCLParameters:
    """Base for classes holding global CCL parameters and their values.

    Subclasses contain a reference to the C-struct with the collection
    of parameters (and their values) (via SWIG). All subclasses act as proxies
    to the CCL parameters at the C-level.

    Subclasses automatically store a backup of the initial parameter state
    to enable ad hoc reloading.
    """

    def __init_subclass__(cls, *, instance=None, factory=None, freeze=False):
        """Routine for subclass initialization.

        Parameters
        ----------
        instance : :obj:`pyccl.ccllib`
            Reference to the instance where the parameters are implemented.
        factory : :class:`pyccl.ccllib`
            The SWIG factory class where the parameters are stored.
        freeze : bool
            Disable parameter mutation.
        """
        super().__init_subclass__()
        if (instance, factory) == (None, None):
            raise ValueError(
                "Provide either the instance, or an instance factory.")
        cls._instance = instance
        cls._factory = factory
        cls._frozen = freeze

    def __init__(self):
        # Emulate abstraction so that base class cannot be instantiated.
        if not (hasattr(self, "_instance") or hasattr(self, "_factory")):
            name = type(self).__name__
            raise TypeError(f"Can't instantiate {name} with no set "
                            "`instance` or `factory`.")
        # Create a new instance if a factory is provided.
        if self._factory:
            object.__setattr__(self, "_instance", self._factory())
        # Keep a copy of the default parameters.
        object.__setattr__(self, "_bak", CCLParameters.get_params_dict(self))

    def __getattribute__(self, name):
        get = object.__getattribute__
        try:
            return get(get(self, "_instance"), name)
        except AttributeError:
            return get(self, name)

    def __setattr__(self, key, value):
        if self._frozen:
            name = self.__class__.__name__
            raise AttributeError(f"Instances of {name} are frozen.")
        if not hasattr(self._instance, key):
            raise KeyError(f"Parameter {key} does not exist.")
        object.__setattr__(self._instance, key, value)

    __getitem__ = __getattribute__

    __setitem__ = __setattr__

    def __repr__(self):
        out = self._bak.copy()
        for par in out:
            out[par] = getattr(self, par)
        return repr(out)

    def reload(self):
        """Reload the C-level default CCL parameters."""
        frozen = self._frozen
        if frozen:
            object.__setattr__(self, "_frozen", False)
        for param, value in self._bak.items():
            setattr(self, param, value)
        object.__setattr__(self, "_frozen", frozen)

    def freeze(self):
        """Freeze an instance of ``CCLParameters``."""
        object.__setattr__(self, "_frozen", True)

    def unfreeze(self):
        """Unfreeze an instance of ``CCLParameters``."""
        object.__setattr__(self, "_frozen", False)

    @classmethod
    def get_params_dict(cls, name):
        """Get a dictionary of the current parameters.

        Arguments:
            name (str or :obj:`CCLParameters`):
                Name or instance of the parameters to look up.
        """
        pars = eval(name) if isinstance(name, str) else name
        out = {}
        for par in dir(pars):
            if not par.startswith("_") and par not in ["this", "thisown"]:
                out[par] = getattr(pars, par)
        return out


class SplineParams(CCLParameters, instance=lib.cvar.user_spline_params):
    """Instances of this class hold the spline parameters."""

    def __setattr__(self, key, value):
        if key == "A_SPLINE_MAX" and value != 1.0:
            raise ValueError("A_SPLINE_MAX is fixed to 1.")
        super().__setattr__(key, value)


class GSLParams(CCLParameters, instance=lib.cvar.user_gsl_params):
    """Instances of this class hold the gsl parameters."""


class PhysicalConstants(CCLParameters, instance=lib.cvar.constants,
                        freeze=True):
    """Instances of this class hold the physical constants."""


spline_params = SplineParams()
gsl_params = GSLParams()
physical_constants = PhysicalConstants()



__all__ = ("ConcentrationConstant",)

import numpy as np

from . import Concentration


class ConcentrationConstant(Concentration):
    """Constant contentration-mass relation.

    Args:
        c (:obj:`float`): constant concentration value.
        mass_def (:class:`~pyccl.halos.massdef.MassDef` or :obj:`str`): a mass
            definition (arbitrary in this case).
    """
    __repr_attrs__ = __eq_attrs__ = ("mass_def", "c",)
    name = 'Constant'

    def __init__(self, c=1, *, mass_def="200c"):
        self.c = c
        super().__init__(mass_def=mass_def)

    def _check_mass_def_strict(self, mass_def):
        return False

    def _concentration(self, cosmo, M, a):
        return np.full_like(M, self.c)[()]



# Taken from github.com/c-d-leonard/IA_GGL/blob/CCL_IAhalo_bm/
# Written by Danielle Leonard.
# This is script is intended to provide code to benchmark the
# integration over the halo mass function and HOD quantities for
# the IA halo model in CCL

import numpy as np
import scipy.integrate
import pyccl as ccl


def normalisation(Mhalo, cosmo, conc, z):
    """ Returns the 'normalisation factor' that goes into the IA halo model
    power spectrum: satelite fraction * Nsat / tot_ns as a function of Mhalo.
    Mhalo is the vector of halo masses over which to integrate in units of Msol
    cosmo is a pyccl cosmology object
    conc is a concentration relation object
    z is the redshift at which to calculate this
    """

    # Get the halo mass function
    HMF_setup = ccl.halos.MassFuncTinker10(cosmo)
    HMF = HMF_setup.get_mass_function(cosmo, Mhalo, 1. / (1. + z))

    # Get HOD quantities we need
    HODHProf = ccl.halos.HaloProfileHOD(conc)
    Ncen_lens = HODHProf._Nc(Mhalo, 1. / (1. + z))
    Nsat_lens = HODHProf._Ns(Mhalo, 1. / (1. + z))

    # Get total number of satellite galaxies
    tot_ns = scipy.integrate.simps(Ncen_lens * Nsat_lens * HMF, np.log10(Mhalo))
    tot_nc = scipy.integrate.simps((Ncen_lens) * HMF, np.log10(Mhalo))

    tot_all = tot_ns + tot_nc

    # Fraction of satellites:
    f_s = tot_ns / tot_all

    return f_s / tot_ns * Nsat_lens


if (__name__ == "__main__"):
    Mhalo = np.logspace(10., 16, 30)  # Units Msol

    z = 0.0

    cosmo = ccl.Cosmology(Omega_c=0.25, Omega_b=0.05, Omega_k=0, sigma8=0.81, n_s=0.96, h=1.)

    concentration = ccl.halos.ConcentrationDuffy08()

    norm_term = normalisation(Mhalo, cosmo, concentration, z)

    save = np.column_stack((Mhalo, norm_term))
    np.savetxt('../IA_halomodel_norm_term.dat', save)





__all__ = ("BaryonsSchneider15",)

import numpy as np

from .. import Pk2D
from . import Baryons


class BaryonsSchneider15(Baryons):
    """The "BCM" model boost factor for baryons. BCM stands for the
    "baryonic correction model" of `Schneider & Teyssier 2015
    <https://arxiv.org/abs/1510.06034>`_. See the

    The boost factor is applied multiplicatively so that
    :math:`P_{\\rm bar.}(k, a) = P_{\\rm DMO}(k, a)\\, f_{\\rm BCM}(k, a)`.

    Args:
        log10Mc (:obj:`float`): logarithmic mass scale of hot
            gas suppression. Defaults to
            :math:`\\log_{10}(1.2\\,10^{14}\\,M_\\odot)`.
        eta_b (:obj:`float`): ratio of escape to ejection radii.
            Defaults to 0.5.
        k_s (:obj:`float`): Characteristic scale (wavenumber) of
            the stellar component in units of :math:`h\\,{\\rm Mpc}^{-1}`.
            Defaults to 55.
    """
    name = 'Schneider15'
    __repr_attrs__ = __eq_attrs__ = ("log10Mc", "eta_b", "k_s")

    def __init__(self, log10Mc=np.log10(1.2E14), eta_b=0.5, k_s=55.0):
        self.log10Mc = log10Mc
        self.eta_b = eta_b
        self.k_s = k_s

    def boost_factor(self, cosmo, k, a):
        """The BCM model boost factor for baryons.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmological parameters.
            k (:obj:`float` or `array`): Wavenumber (in :math:`{\\rm Mpc}^{-1}`).
            a (:obj:`float` or `array`): Scale factor.

        Returns:
            :obj:`float` or `array`: Correction factor to apply to \
            the power spectrum.
        """ # noqa
        a_use, k_use = map(np.atleast_1d, [a, k])
        a_use, k_use = a_use[:, None], k_use[None, :]

        z = 1/a_use - 1
        kh = k_use / cosmo['h']
        b0 = 0.105*self.log10Mc - 1.27
        bfunc = b0 / (1. + (z/2.3)**2.5)
        kg = 0.7 * (1-bfunc)**4 * self.eta_b**(-1.6)
        gf = bfunc / (1 + (kh/kg)**3) + 1. - bfunc
        scomp = 1 + (kh / self.k_s)**2
        fka = gf * scomp

        if np.ndim(k) == 0:
            fka = np.squeeze(fka, axis=-1)
        if np.ndim(a) == 0:
            fka = np.squeeze(fka, axis=0)
        return fka

    def update_parameters(self, log10Mc=None, eta_b=None, k_s=None):
        """Update BCM parameters. All parameters set to ``None`` will
        be left untouched.

        Args:
            log10Mc (:obj:`float`): logarithmic mass scale of hot
                gas suppression.
            eta_b (:obj:`float`): ratio of escape to ejection radii.
            k_s (:obj:`float`): Characteristic scale (wavenumber) of
                the stellar component.
        """
        if log10Mc is not None:
            self.log10Mc = log10Mc
        if eta_b is not None:
            self.eta_b = eta_b
        if k_s is not None:
            self.k_s = k_s

    def _include_baryonic_effects(self, cosmo, pk):
        # Applies boost factor
        a_arr, lk_arr, pk_arr = pk.get_spline_arrays()
        k_arr = np.exp(lk_arr)
        fka = self.boost_factor(cosmo, k_arr, a_arr)
        pk_arr *= fka

        if pk.psp.is_log:
            np.log(pk_arr, out=pk_arr)  # in-place log

        return Pk2D(a_arr=a_arr, lk_arr=lk_arr, pk_arr=pk_arr,
                    is_logp=pk.psp.is_log,
                    extrap_order_lok=pk.extrap_order_lok,
                    extrap_order_hik=pk.extrap_order_hik)



import numpy as np
import pyccl as ccl

import pytest

BEMULIN_TOLERANCE = 1e-3
BEMUNL_TOLERANCE = 5e-3
BEMBAR_TOLERANCE = 1e-3


def test_baccoemu_linear_As_sigma8():
    bemu = ccl.BaccoemuLinear()
    cosmo1 = ccl.Cosmology(
        Omega_c=0.27,
        Omega_b=0.05,
        h=0.67,
        sigma8=0.83,
        n_s=0.96,
        Neff=3.046,
        mass_split='normal',
        m_nu=0.1,
        Omega_g=0,
        Omega_k=0,
        w0=-1,
        wa=0)

    cosmo2 = ccl.Cosmology(
        Omega_c=0.27,
        Omega_b=0.05,
        h=0.67,
        A_s=2.2194e-09,
        n_s=0.96,
        Neff=3.046,
        mass_split='normal',
        m_nu=0.1,
        Omega_g=0,
        Omega_k=0,
        w0=-1,
        wa=0)

    k1, pk1 = bemu.get_pk_at_a(cosmo1, 1.0)
    k2, pk2 = bemu.get_pk_at_a(cosmo2, 1.0)

    err = np.abs(pk1 / pk2 - 1)
    assert np.allclose(err, 0, atol=BEMULIN_TOLERANCE, rtol=0)


def test_baccoemu_nonlinear_As_sigma8():
    bemu = ccl.BaccoemuNonlinear()
    cosmo1 = ccl.Cosmology(
        Omega_c=0.27,
        Omega_b=0.05,
        h=0.67,
        sigma8=0.83,
        n_s=0.96,
        Neff=3.046,
        mass_split='normal',
        m_nu=0.1,
        Omega_g=0,
        Omega_k=0,
        w0=-1,
        wa=0)

    cosmo2 = ccl.Cosmology(
        Omega_c=0.27,
        Omega_b=0.05,
        h=0.67,
        A_s=2.2194e-09,
        n_s=0.96,
        Neff=3.046,
        mass_split='normal',
        m_nu=0.1,
        Omega_g=0,
        Omega_k=0,
        w0=-1,
        wa=0)

    k1, pk1 = bemu.get_pk_at_a(cosmo1, 1.0)
    k2, pk2 = bemu.get_pk_at_a(cosmo2, 1.0)

    err = np.abs(pk1 / pk2 - 1)
    assert np.allclose(err, 0, atol=BEMUNL_TOLERANCE, rtol=0)


def test_baccoemu_baryons_boost():
    baryons = ccl.BaccoemuBaryons()
    nlpkemu = ccl.BaccoemuNonlinear()
    cosmo = ccl.Cosmology(
        Omega_c=0.27,
        Omega_b=0.05,
        h=0.67,
        sigma8=0.83,
        n_s=0.96,
        Neff=3.046,
        mass_split='normal',
        m_nu=0.1,
        Omega_g=0,
        Omega_k=0,
        w0=-1,
        wa=0,
        matter_power_spectrum=nlpkemu)

    k = np.logspace(-2, 0.5, 100)
    cclfk = baryons.boost_factor(cosmo, k, 1)
    pk_gro = cosmo.get_nonlin_power()
    pk_bcm = baryons.include_baryonic_effects(cosmo, pk_gro)
    fk = pk_bcm(k, 1) / pk_gro(k, 1)
    err = np.abs(fk / cclfk - 1)
    assert np.allclose(err, 0, atol=BEMBAR_TOLERANCE, rtol=0)


def test_baccoemu_baryons_changepars():
    baryons = ccl.BaccoemuBaryons()
    baryons.update_parameters(log10_M_c=12.7, log10_eta=-0.4)
    assert ((baryons.bcm_params['M_c'] == 12.7)
            & (baryons.bcm_params['eta'] == -0.4))


def test_baccoemu_baryons_a_range():
    baryons = ccl.BaccoemuBaryons()
    cosmo = ccl.CosmologyVanillaLCDM()
    k = 1e-1
    with pytest.raises(ValueError):
        baryons.boost_factor(cosmo, k, baryons.a_min * 0.9)


def test_baccoemu_baryons_As_sigma8():
    baryons = ccl.BaccoemuBaryons()
    cosmo1 = ccl.Cosmology(
        Omega_c=0.27,
        Omega_b=0.05,
        h=0.67,
        sigma8=0.83,
        n_s=0.96,
        Neff=3.046,
        mass_split='normal',
        m_nu=0.1,
        Omega_g=0,
        Omega_k=0,
        w0=-1,
        wa=0)

    cosmo2 = ccl.Cosmology(
        Omega_c=0.27,
        Omega_b=0.05,
        h=0.67,
        A_s=2.2194e-09,
        n_s=0.96,
        Neff=3.046,
        mass_split='normal',
        m_nu=0.1,
        Omega_g=0,
        Omega_k=0,
        w0=-1,
        wa=0)

    k = np.logspace(-2, 0.5, 100)
    fk1 = baryons.boost_factor(cosmo1, k, 1)
    fk2 = baryons.boost_factor(cosmo2, k, 1)

    err = np.abs(fk1 / fk2 - 1)
    assert np.allclose(err, 0, atol=BEMUNL_TOLERANCE, rtol=0)



__all__ = ("MassFuncWatson13",)

import numpy as np

from . import MassFunc


class MassFuncWatson13(MassFunc):
    """Implements the mass function of `Watson et al. 2013
    <https://arxiv.org/abs/1212.0095>`_. This parametrization accepts
    `fof` and any S.O. masses.

    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef` or :obj:`str`):
            a mass definition object, or a name string.
        mass_def_strict (:obj:`bool`): if ``False``, consistency of the mass
            definition will be ignored.
    """
    name = 'Watson13'

    def __init__(self, *,
                 mass_def="200m",
                 mass_def_strict=True):
        super().__init__(mass_def=mass_def, mass_def_strict=mass_def_strict)

    def _check_mass_def_strict(self, mass_def):
        return mass_def.name == "vir"

    def _get_fsigma_fof(self, cosmo, sigM, a, lnM):
        pA = 0.282
        pa = 2.163
        pb = 1.406
        pc = 1.210
        return pA * ((pb / sigM)**pa + 1.) * np.exp(-pc / sigM**2)

    def _get_fsigma_SO(self, cosmo, sigM, a, lnM):
        om = cosmo.omega_x(a, "matter")
        Delta_178 = self.mass_def.Delta / 178

        if a == 1:
            pA = 0.194
            pa = 1.805
            pb = 2.267
            pc = 1.287
        elif a < 1/(1+6):
            pA = 0.563
            pa = 3.810
            pb = 0.874
            pc = 1.453
        else:
            pA = om * (1.097 * a**3.216 + 0.074)
            pa = om * (5.907 * a**3.058 + 2.349)
            pb = om * (3.136 * a**3.599 + 2.344)
            pc = 1.318

        f_178 = pA * ((pb / sigM)**pa + 1.) * np.exp(-pc / sigM**2)
        C = np.exp(0.023 * (Delta_178 - 1.0))
        d = -0.456 * om - 0.139
        Gamma = (C * Delta_178**d *
                 np.exp(0.072 * (1.0 - Delta_178) / sigM**2.130))
        return f_178 * Gamma

    def _get_fsigma(self, cosmo, sigM, a, lnM):
        if self.mass_def.name == 'fof':
            return self._get_fsigma_fof(cosmo, sigM, a, lnM)
        return self._get_fsigma_SO(cosmo, sigM, a, lnM)



import numpy as np
import pyccl as ccl
from pyccl.modified_gravity import MuSigmaMG

import pytest

# Set tolerances
DISTANCES_TOLERANCE = 1e-4
# The distance tolerance is 1e-3 for distances with massive neutrinos
# This is because we compare to astropy which uses a fitting function
# instead of the full phasespace integral.
# The fitting function itself is not accurate to 1e-4.
DISTANCES_TOLERANCE_MNU = 1e-3

# Tolerance for comparison with CLASS. Currently works up to 1e-6.
DISTANCES_TOLERANCE_CLASS = 1e-5

# Set up the cosmological parameters to be used in each of the models
# Values that are the same for all 5 models
Omega_c = 0.25
Omega_b = 0.05
h = 0.7
A_s = 2.1e-9
n_s = 0.96
Neff = 0.

# Introduce non-zero values of mu_0 and sigma_0 (mu / Sigma
# parameterisation of modified gravity) for some of the tests
mu_0 = 0.1
sigma_0 = 0.1

# Values that are different for the different models
Omega_v_vals = np.array([0.7, 0.7, 0.7, 0.65, 0.75])
w0_vals = np.array([-1.0, -0.9, -0.9, -0.9, -0.9])
wa_vals = np.array([0.0, 0.0, 0.1, 0.1, 0.1])

mnu = [
    [0.04, 0., 0.], [0.05, 0.01, 0.], [0.05, 0., 0.],
    [0.03, 0.02, 0.]]
# For tests with massive neutrinos, we require N_nu_rel + N_nu_mass = 3
# Because we compare with astropy for benchmarks
# Which assumes N total is split equally among all neutrinos.
Neff_mnu = 3.0


def Neff_from_N_ur_N_ncdm(N_ur, N_ncdm):
    """Calculate N_eff from the number of relativistic
    and massive neutrinos."""
    Neff = N_ur + N_ncdm * ccl.DefaultParams.T_ncdm**4 / (4/11)**(4/3)
    return Neff


class_models = {
    "flat_nonu": {
        "Omega_k": 0.0,
        "Neff": 3.0},
    "pos_curv_nonu": {
        "Omega_k": 0.01,
        "Neff": 3.0},
    "neg_curv_nonu": {
        "Omega_k": -0.01,
        "Neff": 3.0},
    "flat_massnu1": {
        "Omega_k": 0.0,
        # 1 massive neutrino
        "Neff": Neff_from_N_ur_N_ncdm(N_ur=2.0, N_ncdm=1.0),
        "m_nu": [0.0, 0.0, 0.1]},
    "flat_massnu2": {
        "Omega_k": 0.0,
        # 3 massive neutrinos
        "Neff": Neff_from_N_ur_N_ncdm(N_ur=0.0, N_ncdm=3.0),
        "m_nu": [0.03, 0.03, 0.1]},
    "flat_massnu3": {
        "Omega_k": 0.0,
        # 3 massive neutrino
        "Neff": Neff_from_N_ur_N_ncdm(N_ur=0.0, N_ncdm=3.0),
        "m_nu": [0.03, 0.05, 0.1]},
    "flat_manynu1": {
        "Omega_k": 0.0,
        "Neff": 6.0},
    "neg_curv_massnu1": {
        "Omega_k": -0.01,
        # 4 massless, 2 massive neutrino
        "Neff": Neff_from_N_ur_N_ncdm(N_ur=4.0, N_ncdm=2.0),
        "m_nu": [0.0, 0.03, 0.1]},
    "pos_curv_manynu1": {
        "Omega_k": 0.01,
        # 3 massless, 3 massive neutrino
        "Neff": Neff_from_N_ur_N_ncdm(N_ur=3.0, N_ncdm=3.0),
        "m_nu": [0.03, 0.05, 0.1]},
    "CCL1": {
        "Omega_k": 0.0,
        "Neff": 3.046},
    "CCL2": {
        "Omega_k": 0.0,
        "w0": -0.9,
        "wa": 0.0,
        "Neff": 3.046},
    "CCL3": {
        "Omega_k": 0.0,
        "w0": -0.9,
        "wa": 0.1,
        "Neff": 3.046},
    "CCL4": {
        "Omega_k": 0.05,
        "w0": -0.9,
        "wa": 0.1,
        "Neff": 3.046},
    "CCL5": {
        "Omega_k": -0.05,
        "w0": -0.9,
        "wa": 0.1,
        "Neff": 3.046},
    "CCL7": {
        "Omega_k": 0.0,
        "Neff": Neff_from_N_ur_N_ncdm(N_ur=2.0, N_ncdm=1.0),
        "m_nu": [0.04, 0.0, 0.0]},
    "CCL8": {
        "Omega_k": 0.0,
        "w0": -0.9,
        "wa": 0.0,
        "Neff": Neff_from_N_ur_N_ncdm(N_ur=1.0, N_ncdm=2.0),
        "m_nu": [0.05, 0.01, 0.0]},
    "CCL9": {
        "Omega_k": 0.0,
        "w0": -0.9,
        "wa": 0.1,
        "Neff": Neff_from_N_ur_N_ncdm(N_ur=0.0, N_ncdm=3.0),
        "m_nu": [0.03, 0.02, 0.04]},
    "CCL10": {
        "Omega_k": 0.05,
        "w0": -0.9,
        "wa": 0.1,
        "Neff": Neff_from_N_ur_N_ncdm(N_ur=2.0, N_ncdm=1.0),
        "m_nu": [0.05, 0.0, 0.0]},
    "CCL11": {
        "Omega_k": -0.05,
        "w0": -0.9,
        "wa": 0.1,
        "Neff": Neff_from_N_ur_N_ncdm(N_ur=1.0, N_ncdm=2.0),
        "m_nu": [0.03, 0.02, 0.0]},
}


def read_chi_benchmark_file():
    """
    Read the file containing all the radial comoving distance benchmarks
    (distances are in Mpc/h)
    """
    # Load data from file
    dat = np.genfromtxt("./benchmarks/data/chi_model1-5.txt").T
    assert (dat.shape == (6, 6))

    # Split into redshift column and chi(z) columns
    z = dat[0]
    chi = dat[1:]
    return z, chi


def read_chi_hiz_benchmark_file():
    """
    Read the file containing all the radial comoving distance benchmarks
    (distances are in Mpc/h)
    """
    # Load data from file
    dat = np.genfromtxt("./benchmarks/data/chi_hiz_model1-3.txt").T
    assert (dat.shape == (4, 7))

    # Split into redshift column and chi(z) columns
    z = dat[0]
    chi = dat[1:]
    return z, chi


def read_chi_mnu_benchmark_file():
    """
    Read the file containing all the radial comoving distance benchmarks
    with non-zero massive neutrinos
    (distances are in Mpc)
    """
    # Load data from file
    dat = np.genfromtxt("./benchmarks/data/chi_mnu_model1-5.txt").T
    assert (dat.shape == (6, 5))

    # Split into redshift column and chi(z) columns
    z = dat[0]
    chi = dat[1:]
    return z, chi


def read_chi_mnu_hiz_benchmark_file():
    """
    Read the file containing all the radial comoving distance benchmarks
    with non-zero massive neutrinos
    (distances are in Mpc)
    """
    # Load data from file
    dat = np.genfromtxt("./benchmarks/data/chi_hiz_mnu_model1-5.txt").T
    assert (dat.shape == (6, 7))

    # Split into redshift column and chi(z) columns
    z = dat[0]
    chi = dat[1:]
    return z, chi


def read_class_allz_chi_benchmark_file():
    """
    Read the file containing the radial comoving distance benchmarks from
    CLASS for the CCL paper models. (distances are in Mpc)
    """
    # Load data from file
    dat = np.genfromtxt("./benchmarks/data/chi_class_allz.txt").T
    assert (dat.shape == (11, 10))

    # Split into redshift column and chi(z) columns
    z = dat[0]
    chi = dat[1:]
    return z, chi


def read_class_mnu_chi_benchmark_file():
    """
    Read the file containing the radial comoving distance benchmarks from
    CLASS additional massive neutrino models. (distances are in Mpc)
    """
    # Load data from file
    dat = np.genfromtxt("./benchmarks/data/chi_class_extra_mnu.txt").T
    assert (dat.shape == (10, 10))

    # Split into redshift column and chi(z) columns
    z = dat[0]
    chi = dat[1:]
    return z, chi


def read_dm_benchmark_file():
    """
    Read the file containing all the distance modulus benchmarks
    """
    # Load data from file
    dat = np.genfromtxt("./benchmarks/data/dm_model1-5.txt").T
    assert (dat.shape == (6, 6))

    # Split into redshift column and chi(z) columns
    z = dat[0]
    dm = dat[1:]
    return z, dm


def read_dm_mnu_benchmark_file():
    """
    Read the file containing all the distance modulus benchmarks
    for non-zero massive neutrinos.
    """
    # Load data from file
    dat = np.genfromtxt("./benchmarks/data/dm_mnu_model1-5.txt").T
    assert (dat.shape == (6, 5))

    # Split into redshift column and chi(z) columns
    z = dat[0]
    dm = dat[1:]
    return z, dm


def read_dm_mnu_hiz_benchmark_file():
    """
    Read the file containing all the distance modulus benchmarks
    for non-zero massive neutrinos at high z.
    """
    # Load data from file
    dat = np.genfromtxt("./benchmarks/data/dm_hiz_mnu_model1-5.txt").T
    assert (dat.shape == (6, 7))

    # Split into redshift column and chi(z) columns
    z = dat[0]
    dm = dat[1:]
    return z, dm


def read_class_allz_dm_benchmark_file():
    """
    Read the file containing the distance modulus benchmarks from
    CLASS for the CCL paper models.
    """
    # Load data from file
    dat = np.genfromtxt("./benchmarks/data/dm_class_allz.txt").T
    assert (dat.shape == (11, 10))

    # Split into redshift column and dm(z) columns
    z = dat[0]
    dm = dat[1:]
    return z, dm


def read_class_mnu_dm_benchmark_file():
    """
    Read the file containing the distance modulus benchmarks from
    CLASS for additional massive neutrino models.
    """
    # Load data from file
    dat = np.genfromtxt("./benchmarks/data/dm_class_extra_mnu.txt").T
    assert (dat.shape == (10, 10))

    # Split into redshift column and dm(z) columns
    z = dat[0]
    dm = dat[1:]
    return z, dm


# Set-up test data
z, chi = read_chi_benchmark_file()
zhi, chi_hiz = read_chi_hiz_benchmark_file()
_, dm = read_dm_benchmark_file()
znu, chi_nu = read_chi_mnu_benchmark_file()
znuhi, chi_nu_hiz = read_chi_mnu_hiz_benchmark_file()
z_class_mnu, chi_class_mnu = read_class_mnu_chi_benchmark_file()
z_dm_class_mnu, dm_class_mnu = read_class_mnu_dm_benchmark_file()
z_class_allz, chi_class_allz = read_class_allz_chi_benchmark_file()
z_dm_class_allz, dm_class_allz = read_class_allz_dm_benchmark_file()
_znu, dm_nu = read_dm_mnu_benchmark_file()
_znuhi, dm_nu_hiz = read_dm_mnu_hiz_benchmark_file()


def compare_distances(z, chi_bench, dm_bench, Omega_v, w0, wa):
    """
    Compare distances calculated by pyccl with the distances in the benchmark
    file.
    This test is only valid when radiation is explicitly set to 0.
    """
    # Set Omega_K in a consistent way
    Omega_k = 1.0 - Omega_c - Omega_b - Omega_v

    cosmo = ccl.Cosmology(
        Omega_c=Omega_c, Omega_b=Omega_b, Neff=Neff,
        h=h, A_s=A_s, n_s=n_s, Omega_k=Omega_k,
        w0=w0, wa=wa, Omega_g=0)

    # Calculate distance using pyccl
    a = 1. / (1. + z)
    chi = ccl.comoving_radial_distance(cosmo, a) * h
    # Compare to benchmark data
    assert np.allclose(chi, chi_bench, atol=1e-12, rtol=DISTANCES_TOLERANCE)

    # compare distance moudli where a!=1
    a_not_one = (a != 1).nonzero()
    dm = ccl.distance_modulus(cosmo, a[a_not_one])

    assert np.allclose(
        dm, dm_bench[a_not_one], atol=1e-3, rtol=DISTANCES_TOLERANCE*10)


def compare_distances_hiz(z, chi_bench, Omega_v, w0, wa):
    """
    Compare distances calculated by pyccl with the distances in the benchmark
    file.
    This test is only valid when radiation is explicitly set to 0.
    """
    # Set Omega_K in a consistent way
    Omega_k = 1.0 - Omega_c - Omega_b - Omega_v

    cosmo = ccl.Cosmology(
        Omega_c=Omega_c, Omega_b=Omega_b, Neff=Neff,
        h=h, A_s=A_s, n_s=n_s, Omega_k=Omega_k,
        w0=w0, wa=wa, Omega_g=0)

    # Calculate distance using pyccl
    a = 1. / (1. + z)
    chi = ccl.comoving_radial_distance(cosmo, a) * h
    # Compare to benchmark data
    assert np.allclose(chi, chi_bench, atol=1e-12, rtol=DISTANCES_TOLERANCE)


def compare_distances_mnu(z, chi_bench, dm_bench, Omega_v, w0, wa, Neff, mnu):
    """
    Compare distances calculated by pyccl with the distances in the benchmark
    file.
    """
    # Set Omega_K in a consistent way
    Omega_k = 1.0 - Omega_c - Omega_b - Omega_v

    cosmo = ccl.Cosmology(
        Omega_c=Omega_c, Omega_b=Omega_b, Neff=Neff_mnu,
        h=h, A_s=A_s, n_s=n_s, Omega_k=Omega_k,
        w0=w0, wa=wa, m_nu=mnu)

    # Calculate distance using pyccl
    a = 1. / (1. + z)
    chi = ccl.comoving_radial_distance(cosmo, a)
    # Compare to benchmark data
    assert np.allclose(chi, chi_bench,
                       atol=1e-12, rtol=DISTANCES_TOLERANCE_MNU)

    # compare distance moudli where a!=1
    a_not_one = (a != 1).nonzero()
    dm = ccl.distance_modulus(cosmo, a[a_not_one])

    assert np.allclose(
        dm, dm_bench[a_not_one], atol=1e-3, rtol=DISTANCES_TOLERANCE_MNU)


def compare_distances_mnu_hiz(z, chi_bench, dm_bench, Omega_v,
                              w0, wa, Neff_mnu, mnu):
    """
    Compare distances calculated by pyccl with the distances in the benchmark
    file.
    """
    # Set Omega_K in a consistent way
    Omega_k = 1.0 - Omega_c - Omega_b - Omega_v

    cosmo = ccl.Cosmology(
        Omega_c=Omega_c, Omega_b=Omega_b, Neff=Neff,
        h=h, A_s=A_s, n_s=n_s, Omega_k=Omega_k,
        w0=w0, wa=wa, m_nu=mnu)

    # Calculate distance using pyccl
    a = 1. / (1. + z)
    chi = ccl.comoving_radial_distance(cosmo, a)
    # Compare to benchmark data
    assert np.allclose(chi, chi_bench,
                       atol=1e-12, rtol=DISTANCES_TOLERANCE_MNU)

    # compare distance moudli where a!=1
    a_not_one = (a != 1).nonzero()
    dm = ccl.distance_modulus(cosmo, a[a_not_one])

    assert np.allclose(
        dm, dm_bench[a_not_one], atol=1e-3, rtol=DISTANCES_TOLERANCE_MNU)


def compare_class_distances(z, chi_bench, dm_bench, Neff=3.0, m_nu=0.0,
                            Omega_k=0.0, w0=-1.0, wa=0.0):
    """
    Compare distances calculated by pyccl with the distances in the CLASS
    benchmark file.
    """
    cosmo = ccl.Cosmology(
        Omega_c=Omega_c, Omega_b=Omega_b, Neff=Neff,
        h=h, A_s=A_s, n_s=n_s, Omega_k=Omega_k, m_nu=m_nu,
        w0=w0, wa=wa)

    # Calculate distance using pyccl
    a = 1. / (1. + z)
    chi = ccl.comoving_radial_distance(cosmo, a)
    # Compare to benchmark data
    assert np.allclose(chi, chi_bench, rtol=DISTANCES_TOLERANCE_CLASS)

    # Compare distance moudli where a!=1
    a_not_one = a != 1
    dm = ccl.distance_modulus(cosmo, a[a_not_one])
    assert np.allclose(dm, dm_bench[a_not_one], rtol=DISTANCES_TOLERANCE_CLASS)


def compare_distances_muSig(z, chi_bench, dm_bench, Omega_v, w0, wa):
    """
    Compare distances calculated by pyccl with the distances in the benchmark
    file, for a ccl cosmology with mu / Sigma parameterisation of gravity.
    Nonzero mu / Sigma should NOT affect distances so we compare to the same
    benchmarks as the mu = Sigma = 0 case deliberately.
    """
    # Set Omega_K in a consistent way
    Omega_k = 1.0 - Omega_c - Omega_b - Omega_v

    # Create new Parameters and Cosmology objects
    cosmo = ccl.Cosmology(
        Omega_c=Omega_c, Omega_b=Omega_b, Neff=Neff,
        h=h, A_s=A_s, n_s=n_s, Omega_k=Omega_k,
        w0=w0, wa=wa, Omega_g=0.,
        mg_parametrization=MuSigmaMG(mu_0=mu_0, sigma_0=sigma_0))

    # Calculate distance using pyccl
    a = 1. / (1. + z)
    chi = ccl.comoving_radial_distance(cosmo, a) * h
    # Compare to benchmark data
    assert np.allclose(chi, chi_bench, atol=1e-12, rtol=DISTANCES_TOLERANCE)

    # compare distance moudli where a!=1
    a_not_one = (a != 1).nonzero()
    dm = ccl.distance_modulus(cosmo, a[a_not_one])

    assert np.allclose(
        dm, dm_bench[a_not_one], atol=1e-3, rtol=DISTANCES_TOLERANCE*10)


def compare_distances_hiz_muSig(z, chi_bench, Omega_v, w0, wa):
    """
    Compare distances calculated by pyccl with the distances in the benchmark
    file, for a ccl cosmology with mu / Sigma parameterisation of gravity.
    Nonzero mu / Sigma should NOT affect distances so we compare to the same
    benchmarks as the mu = Sigma = 0 case deliberately.
    """
    # Set Omega_K in a consistent way
    Omega_k = 1.0 - Omega_c - Omega_b - Omega_v

    # Create new Parameters and Cosmology objects
    cosmo = ccl.Cosmology(
        Omega_c=Omega_c, Omega_b=Omega_b, Neff=Neff,
        h=h, A_s=A_s, n_s=n_s, Omega_k=Omega_k,
        w0=w0, wa=wa, Omega_g=0.,
        mg_parametrization=MuSigmaMG(mu_0=mu_0, sigma_0=sigma_0))

    # Calculate distance using pyccl
    a = 1. / (1. + z)
    chi = ccl.comoving_radial_distance(cosmo, a) * h
    # Compare to benchmark data
    assert np.allclose(chi, chi_bench, atol=1e-12, rtol=DISTANCES_TOLERANCE)


@pytest.mark.parametrize('i', list(range(5)))
def test_distance_model(i):
    compare_distances(
        z, chi[i], dm[i], Omega_v_vals[i], w0_vals[i], wa_vals[i])


@pytest.mark.parametrize('i', list(range(3)))
def test_distance_hiz_model(i):
    compare_distances_hiz(
        zhi, chi_hiz[i], Omega_v_vals[i], w0_vals[i], wa_vals[i])


@pytest.mark.parametrize('i', list(range(4)))
def test_distance_mnu_model(i):
    compare_distances_mnu(
        znu, chi_nu[i], dm_nu[i], Omega_v_vals[i], w0_vals[i],
        wa_vals[i], Neff, mnu[i])


@pytest.mark.parametrize('i', list(range(4)))
def test_distance_mnu_hiz_model(i):
    compare_distances_mnu(znuhi, chi_nu_hiz[i], dm_nu_hiz[i], Omega_v_vals[i],
                          w0_vals[i], wa_vals[i], Neff, mnu[i])


def test_class_distance_model_flat_nonu():
    i = 0
    compare_class_distances(
        z_class_mnu, chi_class_mnu[i], dm_class_mnu[i],
        **class_models["flat_nonu"])


def test_class_distance_model_pos_curv_nonu():
    i = 1
    compare_class_distances(
        z_class_mnu, chi_class_mnu[i], dm_class_mnu[i],
        **class_models["pos_curv_nonu"])


def test_class_distance_model_neg_curv_nonu():
    i = 2
    compare_class_distances(
        z_class_mnu, chi_class_mnu[i], dm_class_mnu[i],
        **class_models["neg_curv_nonu"])


def test_class_distance_model_flat_massnu1():
    i = 3
    compare_class_distances(
        z_class_mnu, chi_class_mnu[i], dm_class_mnu[i],
        **class_models["flat_massnu1"])


def test_class_distance_model_flat_massnu2():
    i = 4
    compare_class_distances(
        z_class_mnu, chi_class_mnu[i], dm_class_mnu[i],
        **class_models["flat_massnu2"])


def test_class_distance_model_flat_massnu3():
    i = 5
    compare_class_distances(
        z_class_mnu, chi_class_mnu[i], dm_class_mnu[i],
        **class_models["flat_massnu3"])


def test_class_distance_model_flat_manynu1():
    i = 6
    compare_class_distances(
        z_class_mnu, chi_class_mnu[i], dm_class_mnu[i],
        **class_models["flat_manynu1"])


def test_class_distance_model_neg_curv_massnu1():
    i = 7
    compare_class_distances(
        z_class_mnu, chi_class_mnu[i], dm_class_mnu[i],
        **class_models["neg_curv_massnu1"])


def test_class_distance_model_pos_curv_massnu1():
    i = 8
    compare_class_distances(
        z_class_mnu, chi_class_mnu[i], dm_class_mnu[i],
        **class_models["pos_curv_manynu1"])


def test_class_allz_distance_model_ccl1():
    i = 0
    compare_class_distances(
        z_class_allz, chi_class_allz[i], dm_class_allz[i],
        **class_models["CCL1"])


def test_class_allz_distance_model_ccl2():
    i = 1
    compare_class_distances(
        z_class_allz, chi_class_allz[i], dm_class_allz[i],
        **class_models["CCL2"])


def test_class_allz_distance_model_ccl3():
    i = 2
    compare_class_distances(
        z_class_allz, chi_class_allz[i], dm_class_allz[i],
        **class_models["CCL3"])


def test_class_allz_distance_model_ccl4():
    i = 3
    compare_class_distances(
        z_class_allz, chi_class_allz[i], dm_class_allz[i],
        **class_models["CCL4"])


def test_class_allz_distance_model_ccl5():
    i = 4
    compare_class_distances(
        z_class_allz, chi_class_allz[i], dm_class_allz[i],
        **class_models["CCL5"])


def test_class_allz_distance_model_ccl7():
    i = 5
    compare_class_distances(
        z_class_allz, chi_class_allz[i], dm_class_allz[i],
        **class_models["CCL7"])


def test_class_allz_distance_model_ccl8():
    i = 6
    compare_class_distances(
        z_class_allz, chi_class_allz[i], dm_class_allz[i],
        **class_models["CCL8"])


def test_class_allz_distance_model_ccl9():
    i = 7
    compare_class_distances(
        z_class_allz, chi_class_allz[i], dm_class_allz[i],
        **class_models["CCL9"])


def test_class_allz_distance_model_ccl10():
    i = 8
    compare_class_distances(
        z_class_allz, chi_class_allz[i], dm_class_allz[i],
        **class_models["CCL10"])


def test_class_allz_distance_model_ccl11():
    i = 9
    compare_class_distances(
        z_class_allz, chi_class_allz[i], dm_class_allz[i],
        **class_models["CCL11"])


@pytest.mark.parametrize('i', list(range(5)))
def test_distance_muSig_model(i):
    compare_distances_muSig(
        z, chi[i], dm[i], Omega_v_vals[i], w0_vals[i], wa_vals[i])


@pytest.mark.parametrize('i', list(range(3)))
def test_distance_hiz_muSig_model(i):
    compare_distances_hiz_muSig(
        zhi, chi_hiz[i], Omega_v_vals[i], w0_vals[i], wa_vals[i])



import os
import numpy as np
import pyccl as ccl
from pyccl.modified_gravity import MuSigmaMG

from scipy.interpolate import interp1d
import pytest


@pytest.fixture(scope='module', params=['fftlog', 'bessel'])
def corr_method(request):
    errfacs = {'fftlog': 0.21, 'bessel': 0.05}
    return request.param, errfacs[request.param]


@pytest.fixture(scope='module')
def set_up(request):
    dirdat = os.path.dirname(__file__) + '/data'
    h0 = 0.67702026367187500
    logA = 3.05  # log(10^10 A_s)
    # scale dependent MG cosmology
    ccl.gsl_params.LENSING_KERNEL_SPLINE_INTEGRATION = False
    ccl.gsl_params.INTEGRATION_LIMBER_EPSREL = 2.5E-5
    ccl.gsl_params.INTEGRATION_EPSREL = 2.5E-5
    cosmo = ccl.Cosmology(Omega_c=0.12/h0**2, Omega_b=0.0221/h0**2, Omega_k=0,
                          h=h0, A_s=np.exp(logA)/10**10, n_s=0.96, Neff=3.046,
                          m_nu=0.0, w0=-1, wa=0, T_CMB=2.7255,
                          mg_parametrization=MuSigmaMG(
                              mu_0=0.1, sigma_0=0.1,
                              c1_mg=1.1, c2_mg=1.1, lambda_mg=1),
                          transfer_function='boltzmann_isitgr',
                          matter_power_spectrum='linear')

    # Ell-dependent correction factors
    # Set up array of ells
    fl = {}
    lmax = 10000
    nls = (lmax - 400)//20+141
    ells = np.zeros(nls)
    ells[:101] = np.arange(101)
    ells[101:121] = ells[100] + (np.arange(20) + 1) * 5
    ells[121:141] = ells[120] + (np.arange(20) + 1) * 10
    ells[141:] = ells[140] + (np.arange(nls - 141) + 1) * 20
    fl['lmax'] = lmax
    fl['ells'] = ells

    # Load dNdz's
    z1, pz1 = np.loadtxt(dirdat + "/bin1_histo.txt", unpack=True)
    z2, pz2 = np.loadtxt(dirdat + "/bin2_histo.txt", unpack=True)

    # Set up the linear galaxy bias as used in generating benchmarks
    bz1 = 1.45*np.ones_like(pz1)
    bz2 = 1.55*np.ones_like(pz2)

    # Initialize tracers
    trc = {}
    trc['g1'] = ccl.NumberCountsTracer(cosmo, has_rsd=False,
                                       dndz=(z1, pz1), bias=(z1, bz1))
    trc['g2'] = ccl.NumberCountsTracer(cosmo, has_rsd=False,
                                       dndz=(z2, pz2), bias=(z2, bz2))
    trc['l1'] = ccl.WeakLensingTracer(cosmo, dndz=(z1, pz1))
    trc['l2'] = ccl.WeakLensingTracer(cosmo, dndz=(z2, pz2))

    # Read benchmarks
    bms = {}
    bms['dd_11'] = np.loadtxt(dirdat+'/wtheta_linear_predictionSD.dat')[0:15]
    bms['dd_22'] = np.loadtxt(dirdat+'/wtheta_linear_predictionSD.dat')[15:30]
    bms['dl_11'] = np.loadtxt(dirdat+'/gammat_linear_predictionSD.dat')[0:15]
    bms['dl_12'] = np.loadtxt(dirdat+'/gammat_linear_predictionSD.dat')[15:30]
    bms['dl_21'] = np.loadtxt(dirdat+'/gammat_linear_predictionSD.dat')[30:45]
    bms['dl_22'] = np.loadtxt(dirdat+'/gammat_linear_predictionSD.dat')[45:60]
    bms['ll_11_p'] = np.loadtxt(dirdat+'/Xip_linear_predictionSD.dat')[0:15]
    bms['ll_12_p'] = np.loadtxt(dirdat+'/Xip_linear_predictionSD.dat')[15:30]
    bms['ll_22_p'] = np.loadtxt(dirdat+'/Xip_linear_predictionSD.dat')[30:45]
    bms['ll_11_m'] = np.loadtxt(dirdat+'/Xim_linear_predictionSD.dat')[0:15]
    bms['ll_12_m'] = np.loadtxt(dirdat+'/Xim_linear_predictionSD.dat')[15:30]
    bms['ll_22_m'] = np.loadtxt(dirdat+'/Xim_linear_predictionSD.dat')[30:45]
    theta = np.loadtxt(dirdat+'/theta_corr_MG.dat')
    bms['theta'] = theta

    # Read error bars
    ers = {}
    d = np.loadtxt("benchmarks/data/sigma_clustering_Nbin5",
                   unpack=True)
    ers['dd_11'] = interp1d(d[0], d[1],
                            fill_value=d[1][0],
                            bounds_error=False)(theta)
    ers['dd_22'] = interp1d(d[0], d[2],
                            fill_value=d[2][0],
                            bounds_error=False)(theta)
    d = np.loadtxt("benchmarks/data/sigma_ggl_Nbin5",
                   unpack=True)
    ers['dl_12'] = interp1d(d[0], d[1],
                            fill_value=d[1][0],
                            bounds_error=False)(theta)
    ers['dl_11'] = interp1d(d[0], d[2],
                            fill_value=d[2][0],
                            bounds_error=False)(theta)
    ers['dl_22'] = interp1d(d[0], d[3],
                            fill_value=d[3][0],
                            bounds_error=False)(theta)
    ers['dl_21'] = interp1d(d[0], d[4],
                            fill_value=d[4][0],
                            bounds_error=False)(theta)
    d = np.loadtxt("benchmarks/data/sigma_xi+_Nbin5",
                   unpack=True)
    # We cut the largest theta angle from xip because of issues
    # with the benchmark.
    ers['ll_11_p'] = interp1d(d[0], d[1],
                              fill_value=d[1][0],
                              bounds_error=False)(theta)
    ers['ll_22_p'] = interp1d(d[0], d[2],
                              fill_value=d[2][0],
                              bounds_error=False)(theta)
    ers['ll_12_p'] = interp1d(d[0], d[3],
                              fill_value=d[3][0],
                              bounds_error=False)(theta)
    d = np.loadtxt("benchmarks/data/sigma_xi-_Nbin5",
                   unpack=True)
    ers['ll_11_m'] = interp1d(d[0], d[1],
                              fill_value=d[1][0],
                              bounds_error=False)(theta)
    ers['ll_22_m'] = interp1d(d[0], d[2],
                              fill_value=d[2][0],
                              bounds_error=False)(theta)
    ers['ll_12_m'] = interp1d(d[0], d[3],
                              fill_value=d[3][0],
                              bounds_error=False)(theta)

    ccl.gsl_params.reload()
    return cosmo, trc, bms, ers, fl


@pytest.mark.parametrize("t1,t2,bm,er,kind,pref",
                         [('g1', 'g1', 'dd_11', 'dd_11', 'NN', 1),
                          ('g2', 'g2', 'dd_22', 'dd_22', 'NN', 1),
                          ('g1', 'l1', 'dl_11', 'dl_11', 'NG', 1),
                          ('g1', 'l2', 'dl_12', 'dl_12', 'NG', 1),
                          ('g2', 'l1', 'dl_21', 'dl_21', 'NG', 1),
                          ('g2', 'l2', 'dl_22', 'dl_22', 'NG', 1),
                          ('l1', 'l1', 'll_11_p', 'll_11_p', 'GG+', 1),
                          ('l1', 'l2', 'll_12_p', 'll_12_p', 'GG+', 1),
                          ('l2', 'l2', 'll_22_p', 'll_22_p', 'GG+', 1),
                          ('l1', 'l1', 'll_11_m', 'll_11_m', 'GG-', 1),
                          ('l1', 'l2', 'll_12_m', 'll_12_m', 'GG-', 1),
                          ('l2', 'l2', 'll_22_m', 'll_22_m', 'GG-', 1)])
def test_xi(set_up, corr_method, t1, t2, bm, er, kind, pref):
    cosmo, trcs, bms, ers, fls = set_up
    method, errfac = corr_method

    # Debugging - define the  same cosmology but in GR

    cl = ccl.angular_cl(cosmo, trcs[t1], trcs[t2], fls['ells'])

    ell = np.arange(fls['lmax'])
    cli = interp1d(fls['ells'], cl, kind='cubic')(ell)
    # Our benchmarks have theta in arcmin
    # but CCL requires it in degrees:
    theta_deg = bms['theta'] / 60.
    xi = ccl.correlation(cosmo, ell=ell, C_ell=cli, theta=theta_deg, type=kind,
                         method=method)
    xi *= pref

    print(xi)

    assert np.all(np.fabs(xi - bms[bm]) < ers[er] * errfac)



import numpy as np
import pytest

import pyccl as ccl
from pyccl.modified_gravity import MuSigmaMG


GROWTH_HIZ_TOLERANCE = 6.0e-6
GROWTH_TOLERANCE = 1e-4

# Set up the cosmological parameters to be used in each of the models
# Values that are the same for all 5 models
Omega_c = 0.25
Omega_b = 0.05
Neff = 0.
m_nu = 0.
h = 0.7
A_s = 2.1e-9
n_s = 0.96

# Values that are different for the different models
Omega_v_vals = np.array([0.7, 0.7, 0.7, 0.65, 0.75, 0.7, 0.7, 0.7, 0.7])
w0_vals = np.array([-1.0, -0.9, -0.9, -0.9, -0.9, -0.9, -0.9, -0.9, -0.9])
wa_vals = np.array([0.0, 0.0, 0.1, 0.1, 0.1, 0.0, 0.0, 0.0, 0.0])
mu0_vals = np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.1, -0.1, 0.1, -0.1])
Sig0_vals = np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.1, -0.1, -0.1, 0.1])


def read_growth_lowz_benchmark_file():
    """
    Read the file containing growth factor benchmarks for the low redshifts.
    """
    # Load data from file
    dat = np.genfromtxt("benchmarks/data/growth_model1-5.txt").T
    assert (dat.shape == (6, 6))

    # Split into redshift column and growth(z) columns
    z = dat[0]
    gfac = dat[1:]
    return z, gfac


def read_growth_highz_benchmark_file():
    """
    Read the file containing growth factor benchmarks for the high redshifts.
    """
    # Load data from file
    dat = np.genfromtxt("benchmarks/data/growth_hiz_model1-3.txt").T
    assert (dat.shape == (4, 7))

    # Split into redshift column and growth(z) columns
    z = dat[0]
    gfac = dat[1:]
    return z, gfac


def read_growth_allz_benchmark_file():
    """
    Read the file containing growth factor benchmarks for
    the whole redshift range.
    """
    # Load data from file
    dat = np.genfromtxt("benchmarks/data/growth_cosmomad_allz.txt").T
    assert (dat.shape == (6, 10))

    # Split into redshift column and growth(z) columns
    z = dat[0]
    gfac = dat[1:]

    return z, gfac


# Set-up test data
z_lowz, gfac_lowz = read_growth_lowz_benchmark_file()
z_highz, gfac_highz = read_growth_highz_benchmark_file()
z_allz, gfac_allz = read_growth_allz_benchmark_file()


def compare_growth(
        z, gfac_bench, Omega_v, w0, wa, mu_0, sigma_0, high_tol=False):
    """
    Compare growth factor calculated by pyccl with the values in the benchmark
    file. This test only works if radiation is explicitly set to 0.
    """

    # Set Omega_K in a consistent way
    Omega_k = 1.0 - Omega_c - Omega_b - Omega_v

    cosmo = ccl.Cosmology(
        Omega_c=Omega_c, Omega_b=Omega_b, Neff=Neff, m_nu=m_nu,
        h=h, A_s=A_s, n_s=n_s, Omega_k=Omega_k, Omega_g=0,
        w0=w0, wa=wa,
        mg_parametrization=MuSigmaMG(mu_0=mu_0, sigma_0=sigma_0))

    # Calculate distance using pyccl
    a = 1. / (1. + z)
    gfac = ccl.growth_factor_unnorm(cosmo, a)

    # Compare to benchmark data
    if high_tol:
        assert np.allclose(gfac, gfac_bench,
                           atol=1e-12, rtol=GROWTH_HIZ_TOLERANCE)
    else:
        assert np.allclose(gfac, gfac_bench,
                           atol=1e-12, rtol=GROWTH_TOLERANCE)


@pytest.mark.parametrize('i', list(range(5)))
def test_growth_lowz_model(i):
    compare_growth(z_lowz, gfac_lowz[i], Omega_v_vals[i], w0_vals[i],
                   wa_vals[i], mu0_vals[i], Sig0_vals[i])


@pytest.mark.parametrize('i', list(range(3)))
def test_growth_highz_model(i):
    compare_growth(z_highz, gfac_highz[i], Omega_v_vals[i], w0_vals[i],
                   wa_vals[i], mu0_vals[i], Sig0_vals[i], high_tol=True)


# 0.01 < z < 1000 tests
@pytest.mark.parametrize('i', list(range(5)))
def test_growth_allz_model(i):
    compare_growth(z_allz, gfac_allz[i], Omega_v_vals[i], w0_vals[i],
                   wa_vals[i], mu0_vals[i], Sig0_vals[i])



import numpy as np
import matplotlib.pyplot as plt

#Generates redshift distribution for angular power spectrum and correlation function benchmarks
#Contact david.alonso@physics.ox.ac.uk if you have issues running this script

nz=256
sz=0.15
z1=1.0
z2=1.5
z1_arr=z1-np.fmin(20*sz,z1)+40*sz*(np.arange(nz)+0.5)/nz
z2_arr=z2-np.fmin(20*sz,z2)+40*sz*(np.arange(nz)+0.5)/nz
pz1_arr=np.exp(-0.5*((z1_arr-z1)/sz)**2)
pz2_arr=np.exp(-0.5*((z2_arr-z2)/sz)**2)

zlo_arr,h1,h2=np.loadtxt("z_DESC-CC",unpack=True)
dz=np.mean(zlo_arr[1:]-zlo_arr[:-1])
zm_arr=zlo_arr+dz/2

z_arr_bias=1.25*(np.arange(2*nz)+0.5)/nz
bias_arr=np.ones(len(z_arr_bias))


plt.plot(z1_arr,pz1_arr)
plt.plot(z2_arr,pz2_arr)
plt.show()

plt.plot(zm_arr,h1)
plt.plot(zm_arr,h2)
plt.show()

plt.plot(z_arr_bias,bias_arr)
plt.show()


np.savetxt("bin1_analytic.txt",np.transpose([z1_arr,pz1_arr]))
np.savetxt("bin2_analytic.txt",np.transpose([z2_arr,pz2_arr]))

np.savetxt("bin1_histo.txt",np.transpose([zm_arr,h1]))
np.savetxt("bin2_histo.txt",np.transpose([zm_arr,h2]))

np.savetxt("bias.txt",np.transpose([z_arr_bias,bias_arr]))



import pyccl
import math
import numpy as np
from scipy.interpolate import interp1d
from astropy.cosmology import Planck15 as cosmo
from astropy.constants import c,G
from astropy import units as u
from scipy.integrate import quad as scipy_int1d

c=c.to(u.km/u.second)

cosmo_fid=dict({'h':0.7,'s8':0.8,'Omc':0.3,'Og':0.0,'Omb':0.0,'Om':0.3,'m_nu':[0,0,0],
                'Neff':3.046,'Omk':0,'tau':0.06,'ns':0.96, 'Tcmb0':0})

# Reset astropy cosmology parameters to cosmo_fid. This is assuming flat LCDM and hence Omega_k=0
cosmo=cosmo.clone(H0=cosmo_fid['h']*100,Ob0=cosmo_fid['Omb'],Om0=cosmo_fid['Om'],
                  m_nu=cosmo_fid['m_nu']*u.eV,Neff=cosmo_fid['Neff'],Tcmb0=cosmo_fid['Tcmb0'])
cosmo_h=cosmo.clone(H0=100)

pk_params={'non_linear':0,'kmax':1E3,'kmin':5.e-5,'nk':5000}

class Power_Spectra():
    def __init__(self,cosmo_params=cosmo_fid,pk_params=pk_params,cosmo=cosmo,cosmo_h=None):
        self.cosmo_params=cosmo_params
        self.pk_params=pk_params
        self.cosmo=cosmo

        if not cosmo_h:
            self.cosmo_h=cosmo.clone(H0=100)
        else:
            self.cosmo_h=cosmo_h

    def Rho_crit(self,cosmo_h=None):
        if not cosmo_h:
            cosmo_h=self.cosmo_h
        G2=G.to(u.Mpc/u.Msun*u.km**2/u.second**2)
        rc=3*self.cosmo_h.H0**2/(8*np.pi*G2)
        rc=rc.to(u.Msun/u.pc**2/u.Mpc)# unit of Msun/pc^2/mpc
        return rc

    def DZ_int(self,z=[0],cosmo=None): #linear growth factor.. full integral.. eq 63 in Lahav and suto
        if not cosmo:
            cosmo=self.cosmo
        def intf(z):
            return (1+z)/(cosmo.H(z).value)**3
        j=0
        Dz=np.zeros_like(z,dtype='float32')

        for i in z:
            Dz[j]=cosmo.H(i).value*scipy_int1d(intf,i,np.inf,epsrel=1.e-6,epsabs=1e-6)[0]
            j=j+1
        Dz*=(2.5*cosmo.Om0*cosmo.H0.value**2)
        return Dz/Dz[0]

    def sigma_crit(self,zl=[],zs=[],cosmo_h=None):
        if not cosmo_h:
            cosmo_h=self.cosmo_h
        ds=cosmo_h.comoving_transverse_distance(zs)
        dl=cosmo_h.comoving_transverse_distance(zl)
        ddls=1-np.multiply.outer(1./ds,dl)#(ds-dl)/ds
        w=(3./2.)*((cosmo_h.H0/c)**2)*(1+zl)*dl/self.Rho_crit(cosmo_h)
        sigma_c=1./(ddls*w)
        x=ddls<=0 #zs<zl
        sigma_c[x]=np.inf
        return sigma_c

    def ccl_pk_bbks(self,z,cosmo_params=None,pk_params=None):
        if not cosmo_params:
            cosmo_params=self.cosmo_params
        if not pk_params:
            pk_params=self.pk_params

        cosmo_ccl=pyccl.Cosmology(h=cosmo_params['h'],Omega_c=cosmo_params['Omc'],Omega_b=cosmo_params['Omb'],
                                  sigma8=cosmo_params['s8'],n_s=cosmo_params['ns'],
                                  transfer_function='bbks', matter_power_spectrum='linear',
                                  Omega_g=cosmo_params['Og'])
                                  #m_nu=cosmo_params['m_nu'],Neff=cosmo_params['Neff'])
        kh=np.logspace(np.log10(pk_params['kmin']),np.log10(pk_params['kmax']),pk_params['nk'])
        nz=len(z)
        ps=np.zeros((nz,pk_params['nk']))
        ps0=[]
        z0=9.#PS(z0) will be rescaled using growth function when CCL fails.

        pyccl_pkf=pyccl.linear_matter_power
        if pk_params['non_linear']==1:
            pyccl_pkf=pyccl.nonlin_matter_power
        for i in np.arange(nz):
            try:
                ps[i]= pyccl_pkf(cosmo_ccl,kh,1./(1+z[i]))
            except Exception as err:
                print ('CCL err',err,z[i])
                if not np.any(ps0):
                    ps0=pyccl.linear_matter_power(cosmo_ccl,kh,1./(1.+z0))
                Dz=self.DZ_int(z=[z0,z[i]])
                ps[i]=ps0*(Dz[1]/Dz[0])**2
        return ps*cosmo_params['h']**3,kh/cosmo_params['h'] #factors of h to get in same units as camb output

    def cl_z(self,z=[],l=np.arange(3000)+1,pk_params=None,cosmo_h=None,
                cosmo=None,pk_func=None):
        if not cosmo_h:
            cosmo_h=self.cosmo_h
        if not pk_func:
            pk_func=self.camb_pk_too_many_z

        nz=len(z)
        nl=len(l)

        pk,kh=pk_func(z=z,pk_params=pk_params)

        cls=np.zeros((nz,nl),dtype='float32')*u.Mpc
        for i in np.arange(nz):
            DC_i=cosmo_h.comoving_transverse_distance(z[i]).value#because camb k in h/mpc
            lz=kh*DC_i-0.5
            DC_i=cosmo_h.comoving_transverse_distance(z[i]).value
            pk_int=interp1d(lz,pk[i]/DC_i**2,bounds_error=False,fill_value=0)
            cls[i][:]+=pk_int(l)*u.Mpc
        return cls

    def kappa_cl(self,zl_min=1.e-4,zl_max=1100,n_zl=10,log_zl=False,pk_func=None,
                zs1=[1100],p_zs1=[1],zs2=[1100],p_zs2=[1],
                pk_params=None,cosmo_h=None,l=np.arange(0,3001)):
        if not cosmo_h:
            cosmo_h=self.cosmo_h

        if log_zl:#bins for z_lens.
            zl=np.logspace(np.log10(max(zl_min,1.e-4)),np.log10(zl_max),n_zl)
        else:
            zl=np.linspace(zl_min,zl_max,n_zl)

        clz=self.cl_z(z=zl,l=l,cosmo_h=cosmo_h,pk_params=pk_params,pk_func=pk_func)
        # P(z,k=(l+05)/chi(z))/chi(z)^2 [nz,nl]
        clz=(clz.T*(c/(cosmo_h.efunc(zl)*cosmo_h.H0))).T
        # P(z,k=(l+0.5)/chi(z))/(chi(z)^2 H(z)) [nz,nl]
        #clz*=(c/(cosmo_h.efunc(zl)*cosmo_h.H0))

        rho=self.Rho_crit(cosmo_h=cosmo_h)*cosmo_h.Om0
        sigma_c1=rho/self.sigma_crit(zl=zl,zs=zs1,cosmo_h=cosmo_h) # 3/2 H0^2 (1+zl) chi(zl) * (1 - chi(zl)/chi(zs)) [n_zs,n_zl]
        sigma_c2=rho/self.sigma_crit(zl=zl,zs=zs2,cosmo_h=cosmo_h)

        dzl=np.gradient(zl)
        dzs1=np.gradient(zs1) if len(zs1)>1 else 1
        dzs2=np.gradient(zs2) if len(zs2)>1 else 1

        cl_zs_12=np.einsum('ji,ki,il',sigma_c2,sigma_c1*dzl,clz)#integrate over zl..
        cl=np.dot(p_zs2*dzs2,np.dot(p_zs1*dzs1,cl_zs_12))
        #Integral[ P(z,k=(l+0.5)/chi(z))/(H(z) chi(z)^2) Integral[3/2 H0^2 (1+z) p(zs1) chi(z) (1-chi(z)/chi(zs1)), dzs1] Integral[3/2 H0^2 (1+zl) chi(zl) (1+chi(zl)/chi(zs2)), dzs2], dz]
        cl/=np.sum(p_zs2*dzs2)*np.sum(p_zs1*dzs1)
        f=l*(l+1.)/(l+0.5)**2 #correction from Kilbinger+ 2017
        cl*=f**2
        #cl*=2./np.pi #comparison with CAMB requires this.
        return l,cl


    def g_kappa_cl(self,pk_func=None,zs=[1100],p_zs=[1],zg=[1],p_zg=[1],bias_g=1,
                pk_params=None,cosmo_h=None,l=np.arange(0,3001)):
        if not cosmo_h:
            cosmo_h=self.cosmo_h

        clz=self.cl_z(z=zg,l=l,cosmo_h=cosmo_h,pk_params=pk_params,pk_func=pk_func) # p(k)/chi**2
        clz*=bias_g

        rho=self.Rho_crit(cosmo_h=cosmo_h)*cosmo_h.Om0
        sigma_c=rho/self.sigma_crit(zl=zg,zs=zs,cosmo_h=cosmo_h)
        dzg=np.gradient(zg) if len(zg)>1 else 1
        dzs=np.gradient(zs) if len(zs)>1 else 1

        cl_zs_12=np.dot(sigma_c*dzg*p_zg,clz)#integrate over zl..
        cl=np.dot(p_zs*dzs,cl_zs_12)
        cl/=np.sum(p_zs*dzs)*np.sum(p_zg*dzg)
        return l,cl


if __name__ == "__main__":

    PS=Power_Spectra()
    #CMB lensing x galaxy clustering
    #Binned 1
    zg=np.genfromtxt('../codecomp_step2_outputs/bin1_histo.txt',
                 names=('z','pz'))
    l,cl_g_cmb=PS.g_kappa_cl(pk_func=PS.ccl_pk_bbks,zs=[1100],p_zs=[1],zg=zg['z'],p_zg=zg['pz'],bias_g=1,)
    np.savetxt("../codecomp_step2_outputs/run_b1b1histo_log_cl_dc.txt",np.column_stack((l,cl_g_cmb)),fmt=['%i','%.18e'])
    
    #Binned 2
    zg=np.genfromtxt('../codecomp_step2_outputs/bin2_histo.txt',
                 names=('z','pz'))
    l,cl_g_cmb=PS.g_kappa_cl(pk_func=PS.ccl_pk_bbks,zs=[1100],p_zs=[1],zg=zg['z'],p_zg=zg['pz'],bias_g=1,)
    np.savetxt("../codecomp_step2_outputs/run_b2b2histo_log_cl_dc.txt",np.column_stack((l,cl_g_cmb)),fmt=['%i','%.18e'])
    
    #Analytic
    sigz1=0.15
    zav=np.linspace(0,10.,10000)
    pza=1./(np.sqrt(2*math.pi)*sigz1)*np.exp(-0.5*((zav-1)/sigz1)**2)
    za={'z':zav,'pz':pza}
    l,cl_g_cmb=PS.g_kappa_cl(pk_func=PS.ccl_pk_bbks,zs=[1100],p_zs=[1],zg=za['z'],p_zg=za['pz'],bias_g=1,)
    np.savetxt("../codecomp_step2_outputs/run_b1b1analytic_log_cl_dc.txt",np.column_stack((l,cl_g_cmb)),fmt=['%i','%.18e'])
    
    #Analytic 2
    pza=1./(np.sqrt(2*math.pi)*sigz1)*np.exp(-0.5*((zav-1.5)/sigz1)**2)
    za={'z':zav,'pz':pza}
    l,cl_g_cmb=PS.g_kappa_cl(pk_func=PS.ccl_pk_bbks,zs=[1100],p_zs=[1],zg=za['z'],p_zg=za['pz'],bias_g=1,)
    np.savetxt("../codecomp_step2_outputs/run_b2b2analytic_log_cl_dc.txt",np.column_stack((l,cl_g_cmb)),fmt=['%i','%.18e'])
    
    #CMB lensing x galaxy lensing (kappa)
    #Binned 1
    zg=np.genfromtxt('../codecomp_step2_outputs/bin1_histo.txt',
                 names=('z','pz'))
    l,cl_g_cmb=PS.kappa_cl(pk_func=PS.ccl_pk_bbks,n_zl=1000,zs2=zg['z'],p_zs2=zg['pz'],zl_max=10)
    np.savetxt("../codecomp_step2_outputs/run_b1b1histo_log_cl_lc.txt",np.column_stack((l,cl_g_cmb)),fmt=['%i','%.18e'])

    #Binned 2
    zg=np.genfromtxt('../codecomp_step2_outputs/bin2_histo.txt',
                     names=('z','pz'))
    l,cl_g_cmb=PS.kappa_cl(pk_func=PS.ccl_pk_bbks,n_zl=1000,zs2=zg['z'],p_zs2=zg['pz'],zl_max=10)
    np.savetxt("../codecomp_step2_outputs/run_b2b2histo_log_cl_lc.txt",np.column_stack((l,cl_g_cmb)),fmt=['%i','%.18e'])

    #Analytic
    sigz1=0.15
    zav=np.linspace(0,10.,10000)
    pza=1./(np.sqrt(2*math.pi)*sigz1)*np.exp(-0.5*((zav-1)/sigz1)**2)
    za={'z':zav,'pz':pza}
    l,cl_g_cmb=PS.kappa_cl(pk_func=PS.ccl_pk_bbks,n_zl=1000,zs2=za['z'],p_zs2=za['pz'],zl_max=10)
    np.savetxt("../codecomp_step2_outputs/run_b1b1analytic_log_cl_lc.txt",np.column_stack((l,cl_g_cmb)),fmt=['%i','%.18e'])
    
    #Analytic 2
    pza=1./(np.sqrt(2*math.pi)*sigz1)*np.exp(-0.5*((zav-1.5)/sigz1)**2)
    za={'z':zav,'pz':pza}
    l,cl_g_cmb=PS.kappa_cl(pk_func=PS.ccl_pk_bbks,n_zl=1000,zs2=za['z'],p_zs2=za['pz'],zl_max=10)
    np.savetxt("../codecomp_step2_outputs/run_b2b2analytic_log_cl_lc.txt",np.column_stack((l,cl_g_cmb)),fmt=['%i','%.18e'])



import numpy as np
import pyccl as ccl
import pytest

EH_TOLERANCE = 1.0e-5


@pytest.mark.parametrize('transfer,fname',
                         [('eisenstein_hu',
                           'model1_pk_eh.txt'),
                          ('eisenstein_hu_nowiggles',
                           'model1_nowig_pk_eh.txt')])
def test_eh(transfer, fname):
    cosmo = ccl.Cosmology(
        Omega_c=0.25,
        Omega_b=0.05,
        h=0.7,
        sigma8=0.8,
        n_s=0.96,
        Neff=0, T_CMB=2.725,
        m_nu=0.0,
        w0=-1.0,
        wa=0.0,
        mass_split='normal',
        Omega_g=0,
        Omega_k=0,
        transfer_function=transfer,
        matter_power_spectrum='linear')

    data = np.loadtxt('./benchmarks/data/' + fname)

    k = data[:, 0] * cosmo['h']
    for i in range(1):
        a = 1.0 / (1.0 + i)
        pk = data[:, i+1] / (cosmo['h']**3)
        pk_ccl = ccl.linear_matter_power(cosmo, k, a)
        err = np.abs(pk_ccl/pk - 1)
        assert np.allclose(err, 0, rtol=0, atol=EH_TOLERANCE)
